"""Functions to calculate energy requirement calculations.

This module provides functions to estimate the energy requirements for maintenance, 
growth, lactation, and gestation.
"""

def calculate_An_NEmUse_NS(
    An_StatePhys: str, 
    An_BW: float, 
    An_BW_empty: float,
    An_Parity_rl: int, 
    Dt_DMIn_ClfLiq: float
) -> float:
    """
    *Calculate Animal (An) Net Energy (NE) of maintenance in unstressed (NS) dairy cows*

    Calculate the NE required for maintenance in unstressed dairy cows, taking into account the physiological state, 
    body weight, empty body weight, parity, and dry matter intake from calf liquid diet.

    Parameters
    ----------
    An_StatePhys : str
        The physiological state of the animal ("Calf", "Heifer", "Dry Cow", "Lactating Cow", "Other").
    An_BW : float
        The body weight of the animal in kg.
    An_BW_empty : float
        The empty body weight of the animal in kg, applicable for calves.
    An_parity_rl : int
        The parity of the cow as real value from 0 to 2.
    Dt_DMIn_ClfLiq : float
        The dry matter intake from calf liquid diet in kg, applicable for calves.

    Returns
    -------
    float
        The net energy required for maintenance (NEm) in unstressed cows, in megacalories per day (mcal/d).

    Notes
    -----
    - The calculation varies based on the physiological state of the animal, with specific adjustments for calves on milk or mixed diet,
      weaned calves, heifers, and cows.
    - Reference to specific lines in the Nutrient Requirements of Dairy Cattle R Code:
        - Heifers: Line 2777
        - Calves on milk or mixed diet: Line 2779
        - Weaned calves: Line 2780
        - Cows: Line 2782
    - Based on following equations from Nutrient Requirements of Dairy Cattle book:
        - An_NEmUse_NS for cow and heifer is same as NELmaint (Mcal/d) from Equation 3-13
        - km = 0.0769 (milk and/or milk + solid) & km = 0.97 (weaned calves) from Table 10-1 (Item: NEm, kcal/kg EBW^0.75)
        - __NOTE__: these are not consistent with the values presented in Equation 20-272 ?


    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example for a calf on a milk diet
    nd.calculate_An_NEmUse_NS(
        An_StatePhys='Calf',
        An_BW=93,
        An_BW_empty=85,
        An_Parity_rl=0,
        Dt_DMIn_ClfLiq=4
    )
    ```
    # """

    # Heifers, R code line 2777
    # R code note: 'Back calculated from MEm of 0.15 and Km_NE_ME = 0.66'
    An_NEmUse_NS = 0.10 * An_BW**0.75
    # Calves drinking milk or eating mixed diet
    # R code line 2779
    if An_StatePhys == "Calf" and Dt_DMIn_ClfLiq > 0:
        An_NEmUse_NS = 0.0769 * An_BW_empty**0.75
    # Adjust NEm for weaned calves (Calf state with zero DMI from calf liquid diet)
    # R code line 2780
    elif An_StatePhys == "Calf" and Dt_DMIn_ClfLiq == 0:
        An_NEmUse_NS = 0.097 * An_BW_empty**0.75
    # Adjust NEm for cows based on parity (assuming parity > 0 implies cow)
    # R code line 2782
    elif An_Parity_rl > 0:
        # This recalculates what is already set as default for Heifers
        # Equation 20-272 says An_BW is An_BW NPr_3 i.e. Equation 20-246
        An_NEmUse_NS = 0.10 * An_BW**0.75
    return An_NEmUse_NS


def calculate_An_NEm_Act_Graze(
    Dt_PastIn: float, 
    Dt_DMIn: float,
    Dt_PastSupplIn: float, 
    An_MBW: float
) -> float:
    """
    *Calculate Animal (An) Net Energy (NE) of maintenance for grazing activity (Act_Graze)*
    
    Calculate the NE used for grazing activity (NEm_Act_Graze).
    This function estimates the additional energy expenditure due to grazing based on pasture intake, total dry matter intake, pasture supplementation,
    and the metabolic body weight of the animal.

    Parameters
    ----------
    Dt_PastIn : float
        The dry matter intake from pasture in kg/d.
    Dt_DMIn : float
        The total dry matter intake in kg/d.
    Dt_PastSupplIn : float
        The dry matter intake from supplementation in kg/d. E.g. could be supplemental concentrate or forage
    An_MBW : float
        The metabolic body weight of the animal in kg, typically calculated as the live body weight in kg to the power of 0.75.

    Returns
    -------
    float
        The net energy used for grazing activity, in Mcal/d.

    Notes
    -----
    - The energy expenditure for grazing activity is considered only if the proportion of pasture intake to total dry matter intake is above a certain threshold.
    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code:
        - Lines 2793-2795
    - Based on following equations from Nutrient Requirements of Dairy Cattle book:
        - Equation 20-274
    - This function assumes a linear decrease in energy expenditure for grazing as pasture supplementation increases, with a base value adjusted for the metabolic body weight of the animal.

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of NE used for grazing activity
    nd.calculate_An_NEm_Act_Graze(
        Dt_PastIn=15,
        Dt_DMIn=20,
        Dt_PastSupplIn=5,
        An_MBW=500
    )
    ```
    """
    if Dt_PastIn / Dt_DMIn < 0.005:  # Line 2793
        An_NEm_Act_Graze = 0
    else:
        An_NEm_Act_Graze = 0.0075 * An_MBW * (600 - 12 * Dt_PastSupplIn) / 600  
        # Lines 2794-5
    return An_NEm_Act_Graze


def calculate_An_NEm_Act_Parlor(
    An_BW: float, 
    Env_DistParlor: float,
    Env_TripsParlor: float
) -> float:
    """
    *Calculate Animal (An) Net Energy (NE) of maintenance for activity walk to/from parlor (Act_Parlor)*
    
    Calculate the NE used for activity related to walking to and from the parlor (NEm_Act_Parlor).
    This function estimates the NE expenditure based on the distance from barn or paddock to the parlor,
    the number of daily trips to and from the parlor, and the body weight of the animal.

    Parameters
    ----------
    An_BW : float
        The body weight of the animal in kg.
    Env_DistParlor : float
        The distance from barn or paddock to the parlor in meters.
    Env_TripsParlor : float
        The number of daily trips to and from the parlor, usually two times the number of milkings.

    Returns
    -------
    float
        The net energy used for walking to the parlor activity, in Mcal/d.

    Notes
    -----
    - The energy expenditure for walking to the parlor is calculated by considering the distance, the number of trips,
      and the animal's body weight to estimate the total energy spent in this activity.
    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code: Line 2796
    - Based on following equations from Nutrient Requirements of Dairy Cattle book: Equation 20-275

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of NE used for walking to the parlor
    nd.calculate_An_NEm_Act_Parlor(
        An_BW=650,
        Env_DistParlor=500,
        Env_TripsParlor=4
    )
    ```
    """
    An_NEm_Act_Parlor = (0.00035 * Env_DistParlor /
                         1000) * Env_TripsParlor * An_BW  # Line 2796
    return An_NEm_Act_Parlor


def calculate_An_NEm_Act_Topo(An_BW: float, Env_Topo: float) -> float:
    """
    *Calculate Animal (An) Net Energy (NE) of maintenance for topographical activity (Act_Topo)*
    
    Calculate the NE used due to topographical activity (NEm_Act_Topo) in dairy cows, measured in Megacalories per day (Mcal/d).
    This function estimates the energy expenditure associated with the daily total climb due to topography and during transit between
    the milking parlor and the barn or paddock, considering only the meters of uphill climb, as downhill movement has negligible locomotion cost.

    Parameters
    ----------
    An_BW : float
        The body weight of the animal in kg.
    Env_Topo : float
        The positive elevation change per day in meters, considering only the uphill climb.

    Returns
    -------
    float
        The net energy used due to topography, in Megacalories per day (Mcal/d).

    Notes
    -----
    - The calculation takes into account the physical effort required to overcome positive elevation changes while grazing and in transit
      to and from milking, which is directly proportional to the animal's body weight and the total meters climbed uphill.
    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code:
        - Main calculation: Line 2797
    - This calculation is based on following equations from the Nutrient Requirements of Dairy Cattle book: Equation 20-276

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of NE used due to topographical activity
    nd.calculate_An_NEm_Act_Topo(
        An_BW=650,
        Env_Topo=100  # Assuming 100 meters of uphill climb per day
    )
    ```
    """
    An_NEm_Act_Topo = 0.0067 * Env_Topo / 1000 * An_BW  # Line 2797
    return An_NEm_Act_Topo


def calculate_An_NEmUse_Act(
    An_NEm_Act_Graze: float, 
    An_NEm_Act_Parlor: float,
    An_NEm_Act_Topo: float
) -> float:
    """
    *Calculate Animal (An) Net Energy (NE) of maintenance use (mUse) for activity (Act)*
    
    Calculate the total NE used for activity (An_NEmUse_Act) in dairy cows, measured in Megacalories per day (Mcal/d).
    This function sums the energy expenditures due to grazing, walking to and from the parlor, and navigating topography.

    Parameters
    ----------
    An_NEm_Act_Graze : float
        The net energy used for grazing activity, in Mcal/d. Normally calculated by [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_An_NEm_Act_Graze`).
    An_NEm_Act_Parlor : float
        The net energy used for walking to and from the parlor, in Mcal/d. Normally calculated by [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_An_NEm_Act_Parlor`).
    An_NEm_Act_Topo : float
        The net energy used due to topography (positive elevation change), in Mcal/d. Normally calculated by [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_An_NEm_Act_Topo`).

    Returns
    -------
    float
        The total net energy used for activity, in Megacalories per day (Mcal/d).

    Notes
    -----
    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code:
        - Main calculation: Line 2798
    - This calculation is based on following equations from the Nutrient Requirements of Dairy Cattle book: Equation 20-277

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of total NE used for activity
    nd.calculate_An_NEmUse_Act(
        An_NEm_Act_Graze=3.3,
        An_NEm_Act_Parlor=0.5,
        An_NEm_Act_Topo=0.4
    )
    ```
    """
    An_NEmUse_Act = An_NEm_Act_Graze + An_NEm_Act_Parlor + An_NEm_Act_Topo 
    # Line 2798
    return An_NEmUse_Act


def calculate_An_NEmUse(
    An_NEmUse_NS: float, 
    An_NEmUse_Act: float,
    coeff_dict: dict
) -> float:
    """
    *Calculate Animal (An) Net Energy (ME) maintenance use (mUse)*

    Calculate the total net energy (NE) used for maintenance (NEmUse) in dairy cows.
    This function computes the sum of net energy for non-stressed maintenance, net energy used for activities, and net energy used due to environmental factors.

    Parameters
    ----------
    An_NEmUse_NS : float
        The net energy used for non-stressed maintenance, in Mcal/d. Normally calculated by [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_An_NEmUse_NS`).
    An_NEmUse_Act : float
        The total net energy used for activity, in Mcal/d. Normally calculated by [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_An_NEmUse_Act`).
    coeff_dict : dict
        A dictionary containing coefficients required for the calculation, specifically 'An_NEmUse_Env' for environmental energy use.

    Returns
    -------
    float
        The total net energy used for maintenance, in Megacalories per day (Mcal/d).

    Notes
    -----
    - Requires the coefficient for environmental net energy use ('An_NEmUse_Env') within the coeff_dict
    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code:
        - Main calculation: Line 2802
    - This calculation is based on following equations from the Nutrient Requirements of Dairy Cattle book: Equation 20-278

    - TODO: The An_NEmUse_Env should be 0 unless otherwise calculated, lines 2786-2790. Consider removing from coeff_dict.

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of total NE used for maintenance
    example_coeff_dict = {'An_NEmUse_Env': 0.05}
    nd.calculate_An_NEmUse(
        An_NEmUse_NS=1.2,
        An_NEmUse_Act=0.35,
        coeff_dict=example_coeff_dict
    )
    ```
    """
    An_NEmUse = An_NEmUse_NS + coeff_dict['An_NEmUse_Env'] + An_NEmUse_Act  
    # Line 2802
    return An_NEmUse


def calculate_An_MEmUse(An_NEmUse: float, Km_ME_NE: float) -> float:
    """
    *Calculate Animal (An) Metabolizable Energy (ME) maintenance use (mUse)*

    This function converts net energy (NE) use for maintenance into ME use by applying a conversion efficiency coefficient (Km_ME_NE),
    which varies based on physiological state and specific feeding conditions.

    Parameters
    ----------
    An_NEmUse : float
        The total net energy used for maintenance, in Mcal/d. Normally calculated by [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_An_NEmUse`).
    coeff_dict : dict
        A dictionary containing the conversion coefficient 'Km_ME_NE' for converting NEm to MEm.

    Returns
    -------
    float
        The total metabolizable energy used for maintenance, in Megacalories per day (Mcal/d).

    Notes
    -----
    - The conversion coefficient 'Km_ME_NE' is determined based on the animal's physiological state
      and specific feeding conditions, such as dry feed only, liquid feed only, or mixed feeding for calves, and standard values for heifers,
      lactating cows, or dry cows.
    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code:
        - Main calculation: Line 2845
     - This calculation is based on following equations from the Nutrient Requirements of Dairy Cattle book: Equations 20-279 to 20-282

    - TODO: Remove coeff_dict and replace with further functions, Include refactored code below for the original lines 2806-2818 in R Code.
    - NOTE: The value used for liquid feed in R is 0.723 but in book it is 0.718 (equation 20-280)

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of total MEm used for maintenance
    coeff_dict = {'Km_ME_NE': 0.66}  # Example coefficient for a lactating cow
    nd.calculate_An_MEmUse(
        An_NEmUse=2.5,
        coeff_dict=coeff_dict
    )
    ```
    """
    An_MEmUse = An_NEmUse / Km_ME_NE  # Line 2845
    return An_MEmUse


def calculate_Rsrv_NEgain(Rsrv_Fatgain: float, Rsrv_CPgain: float) -> float:
    """
    *Calculate Reserve (Rsrv) Net Energy (NE) Gain*

    This function estimates the NE content of body reserve gains based on the relative proportions of fat and protein gained,
    using their respective heats of combustion.

    Parameters
    ----------
    Rsrv_Fatgain : float
        The amount of body fat gained in kg/day. Normally calculated by [](`~nasem_dairy.nasem_equations.body_composition.calculate_Rsrv_Fatgain`).
    Rsrv_CPgain : float
        The amount of crude protein (CP) gained in kg/day. Normally calculated by [](`~nasem_dairy.nasem_equations.body_composition.calculate_Rsrv_CPgain`).

    Returns
    -------
    float
        The net energy of body reserve gain, in Megacalories per day (Mcal/d).

    Notes
    -----
    - The energy value of a kilogram of true body tissue that is lost or gained is dependent on the relative proportions of fat and protein in the tissue and their respective heat of combustion.
    - The committee, as in the seventh edition of the Nutrient Requirements of Dairy Cattle, chose 9.4 Mcal/kg for retained body fat and 5.55 Mcal/kg for retained body protein as the heats of combustion.
    - This calculation is based on following equations from the Nutrient Requirements of Dairy Cattle book: Equation 3-20c
    
    - NOTE: Comment in original R code: "These are really REgain.  Abbreviations on NEgain need to be sorted out. MDH"

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of NE of body reserve gain
    nd.calculate_Rsrv_NEgain(
        Rsrv_Fatgain=0.2,  # 0.2 kg of body fat gained per day
        Rsrv_CPgain=0.05   # 0.05 kg of crude protein gained per day
    )
    ```
    """
    Rsrv_NEgain = 9.4 * Rsrv_Fatgain + 5.55 * Rsrv_CPgain
    # Line 2867 #These are really REgain. Abbreviations on NEgain need to be
    # sorted out. MDH
    return Rsrv_NEgain


def calculate_Kr_ME_RE(Trg_MilkProd: float, Trg_RsrvGain: float) -> float:
    """
    *Calculate Kr (efficiency of converting) Metabolizable Energy (ME) to Retained Energy (RE)*

    Select the efficiency of ME conversion to RE. The efficiency varies based on the physiological state and whether
    the animal is gaining or losing body reserves.

    Parameters
    ----------
    Trg_MilkProd : float
        The targeted milk production in kg/day. A value greater than 0 indicates a lactating cow.
    Trg_RsrvGain : float
        The targeted body reserve gain in kg/day. Positive values indicate gain, while negative values indicate loss.

    Returns
    -------
    float
        The efficiency of ME to RE conversion for reserves gain or loss, varying by physiological state.

    Notes
    -----
    - This function categorizes animals into three scenarios based on their milk production and reserve gain targets:
        1. Lactating cows with positive reserve gain have an efficiency of 0.75.
        2. Animals (typically cows) losing reserves have an efficiency of 0.89, reflecting the higher efficiency when metabolizing reserves for energy.
        3. Heifers and dry cows gaining reserves, or lactating cows without specified milk production but with reserve gain, have an efficiency of 0.60.

    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code:
        - Lines 2835-7
    - Values from book appear to be from Table 3-2.
        - Text on Page 34 (above equation 3-19a) & Table 3-2 mentions a Kr_ME_RE of 0.74 for lactating cows (but 0.75 is used in model)
        - Dry cow: 0.60 is consistent between model and Table 3-2
        - 0.89 appears to be the conversion of RE to NEL, i.e. because gain is negative the RE is being converted to NEL

    
    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation for a lactating cow gaining reserves
    nd.calculate_Kr_ME_RE(
        Trg_MilkProd=20,  # 20 kg of milk production per day
        Trg_RsrvGain=0.2  # 0.2 kg of body reserve gain per day
    )
    ```

    """
    if Trg_MilkProd > 0 and Trg_RsrvGain > 0:  
        # Efficiency of ME to Rsrv RE for lactating cows gaining Rsrv, Line 2836
        Kr_ME_RE = 0.75
    elif Trg_RsrvGain <= 0:  
        # Line 2837, Efficiency of ME generated for cows losing Rsrv
        Kr_ME_RE = 0.89
    else:
        # Efficiency of ME to RE for reserves gain, Heifers and dry cows, Line 2835
        Kr_ME_RE = 0.60
    return Kr_ME_RE


def calculate_Rsrv_MEgain(Rsrv_NEgain: float, Kr_ME_RE: float) -> float:
    """
    *Calculate Reserve (Rsrv) Metabolizable Energy (ME) Gain*

    Calculate the metabolizable energy (ME) of body reserve gain in dairy cows, measured in Megacalories per day (Mcal/d).
    This function uses the net energy (NE) gain from body reserves and the efficiency of ME conversion to reserve energy (RE) to
    estimate the total ME associated with body reserve gains.

    Parameters
    ----------
    Rsrv_NEgain : float
        The net energy of body reserve gain, in Mcal/d. Normally calculated by [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_Rsrv_NEgain`).
    Kr_ME_RE : float
        The efficiency of converting ME to RE for reserves gain. Normally calculated by [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_Kr_ME_RE`).

    Returns
    -------
    float
        The metabolizable energy associated with body reserve gain, in Megacalories per day (Mcal/d).

    Notes
    -----
    - The calculation inversely applies the conversion efficiency (Kr_ME_RE) to the net energy gain (Rsrv_NEgain) to estimate the metabolizable
      energy required for the observed gain in body reserves.
    - Reference to line number in the Nutrient Requirements of Dairy Cattle R Code: Line 2872

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of ME associated with body reserve gain
    nd.calculate_Rsrv_MEgain(
        Rsrv_NEgain=1.5,  # 1.5 Mcal/d of net energy gain from body reserves
        Kr_ME_RE=0.75     # Efficiency of ME to RE conversion for reserves gain
    )
    ```
    """
    Rsrv_MEgain = Rsrv_NEgain / Kr_ME_RE  # Line 2872
    return Rsrv_MEgain


def calculate_Frm_NEgain(Frm_Fatgain: float, Frm_CPgain: float) -> float:
    """
    *Calculate Frame (Frm) Net Energy (NE) Gain*

    Calculate the net energy (NE) of frame gain in dairy cows, measured in Megacalories per day (Mcal/d).
    This function estimates the NE content of frame gains based on the amounts of fat and protein gained,
    utilizing their respective heats of combustion.

    Parameters
    ----------
    Frm_Fatgain : float
        The amount of fat gained in the frame, in kg/d. Normally calculated by [](`~nasem_dairy.nasem_equations.body_composition.calculate_Frm_Fatgain`).
    Frm_CPgain : float
        The amount of crude protein (CP) gained in the frame, in kg/d. Normally calculated by [](`~nasem_dairy.nasem_equations.body_composition.calculate_Frm_CPgain`).

    Returns
    -------
    float
        The retained energy of frame gain, in Mcal/d.

    Notes
    -----
    - The energy value of frame tissue gain is calculated similarly to body reserve gains, based on the relative proportions of fat and protein in the tissue and their respective heat of combustion.
    - The committee, as in the seventh edition of the Nutrient Requirements of Dairy Cattle, chose 9.4 Mcal/kg for fat and 5.55 Mcal/kg for protein as the heats of combustion for these components (page 257)
    - This calculation is based on following equations from the Nutrient Requirements of Dairy Cattle book:
       - Equation 11-5d (RE_EBG) - However, this version takes Frm_Fatgain and Frm_CPgain which are kg/d (based on target frame gain input) whereas the book calculates per kg of gain, i.e. with Fat_EBG (kg/kg) and Protein_EBG (kg/kg)
       - Note that Equation 3-20c (RE_FADG; frame average daily gain) is similar, but is correcting for empty body weight in 3-20a and 2-20b via (EBG/ADG) which is empty body weight gain / live BW gain, which is 0.85 for 15% assumption.
    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code: Line 2867.

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of NE of frame gain
    nd.calculate_Frm_NEgain(
        Frm_Fatgain=0.1,  # 0.1 kg of fat gained in the frame per day
        Frm_CPgain=0.05   # 0.05 kg of crude protein gained in the frame per day
    )
    ```
    """
    Frm_NEgain = 9.4 * Frm_Fatgain + 5.55 * Frm_CPgain  # Line 2867
    return Frm_NEgain


def calculate_Frm_MEgain(Frm_NEgain: float, Kf_ME_RE: float) -> float:
    """
    *Calculate Frame (Frm) Metabolizable Energy (ME) Gain*

    This function converts retained energy (RE) gain from frame tissue into ME using a conversion coefficient (Kf_ME_RE).

    Parameters
    ----------
    Frm_NEgain : float
        The net energy gain from frame tissue (should be called retained energy), in Mcal/d. Normally calculated by [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_Frm_NEgain`)
    coeff_dict : dict
        A dictionary containing the conversion coefficient 'Kf_ME_RE' for converting RE to ME for frame gains.

    Returns
    -------
    float
        The metabolizable energy associated with frame gain, in Megacalories per day (Mcal/d).

    Notes
    -----
    - The conversion coefficient 'Kf_ME_RE' is currently specified in the coeff_dict. 
    - TODO: However, update code with logic for selecting correct Kf_ME_RE - might still need to have defaults set in coeff_dict for calf liquid (0.56) and cow/heifer (0.4)

    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code: Line 2873.
    - This calculation is based on following equations from the Nutrient Requirements of Dairy Cattle book:
        - Equation 3-20d - noting that here `Frm_NEgain` should really be called RE (retained energy), and that 0.4 in book is default value for Kf_ME_RE

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of ME associated with frame gain
    coeff_dict = {'Kf_ME_RE': 0.4}  # Example conversion coefficient for frame gain
    nd.calculate_Frm_MEgain(
        Frm_NEgain=1.0,  # 1.0 Mcal/d of net energy gain from frame tissue
        coeff_dict=coeff_dict
    )
    ```
    """
    Frm_MEgain = Frm_NEgain / Kf_ME_RE  # Line 2873
    return Frm_MEgain


def calculate_Kf_ME_RE_ClfDry(An_DE: float) -> float:
    Kf_ME_RE_ClfDry = ((1.1376 * An_DE * 0.93 - 0.1198 * (An_DE * 0.93)**2 + 
                        0.0076 * (An_DE * 0.93)**3 - 1.2979) / (An_DE * 0.93))
    return Kf_ME_RE_ClfDry 


def calculate_Kf_ME_RE(
    An_StatePhys: str, 
    Kf_ME_RE_ClfDry: float, 
    Dt_DMIn_ClfLiq: float, 
    Dt_DMIn: float,
    coeff_dict: dict
) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {"Kf_ME_RE_ClfLiq": 0.56}
    
    calculate_Kf_ME_RE(
        An_StatePhys = "Calf", Kf_ME_RE_ClfDry = 0.5, Dt_DMIn_ClfLiq = 2.0, 
        Dt_DMIn = 5.0, coeff_dict = coeff_dict
    )
    ```
    """
    if An_StatePhys == "Calf": # Equation 20-223
        Kf_ME_RE = (coeff_dict['Kf_ME_RE_ClfLiq'] * Dt_DMIn_ClfLiq / Dt_DMIn + 
                    Kf_ME_RE_ClfDry * (Dt_DMIn - Dt_DMIn_ClfLiq) / Dt_DMIn)
    else:
        Kf_ME_RE = 0.4
    return Kf_ME_RE


def calculate_An_MEgain(Rsrv_MEgain: float, Frm_MEgain: float) -> float:
    """
    *Calculate Animal (An) Metabolizable Energy (ME) Gain*

    Calculate the total ME required for both frame and reserve gain in the animal, measured in Megacalories per day (Mcal/d).

    Parameters
    ----------
    Rsrv_MEgain : float
        The metabolizable energy of body reserve gain, in Mcal/d. Normally calculated by 
        [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_Rsrv_MEgain`).
    Frm_MEgain : float
        The metabolizable energy of frame gain, in Mcal/d. Normally calculated by 
        [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_Frm_MEgain`).

    Returns
    -------
    float
        The total metabolizable energy required for frame and reserve gain, in Mcal/d.

    Notes
    -----
    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code: Line 2874.
    - Equation from the Nutrient Requirements of Dairy Cattle book: Equation 20-247

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of total ME required for frame and reserve gain
    nd.calculate_An_MEgain(
        Rsrv_MEgain=1.5,  # Metabolizable energy gain from body reserves, in Mcal/d
        Frm_MEgain=0.5    # Metabolizable energy gain from frame, in Mcal/d
    )
    ```
    """
    An_MEgain = Rsrv_MEgain + Frm_MEgain  # Line 2874
    return An_MEgain


def calculate_Gest_REgain(GrUter_BWgain: float, coeff_dict: dict) -> float:
    """
    *Calculate Gestation (Gest) Retained Energy (RE) Gain*

    Calculate the gestational requirements for net energy (NE, Mcal/d) based on the change in gravid uterus weight during gestation (GrUter_BWgain; kg/d) 
    multiplied by the concentration of energy in the final gravid uterus at parturition (NE_GrUtWt, Mcal/kg)
    

    Parameters
    ----------
    GrUter_BWgain : float
        The gross uterine body weight gain (or loss) in kg/d, representing changes due to uterine growth or regression. Usually calculated by [](`~nasem_dairy.nasem_equations.gestation_equations.calculate_GrUter_BWgain`)
    coeff_dict : dict
        A dictionary containing the coefficient 'NE_GrUtWt' which is the concentration of NE per kg of fresh Gravid Uterus weight at birth (Mcal/kg), defaults to 0.95.

    Returns
    -------
    float
        The RE gain for gestation (called NE in book), accounting for uterine growth or regression, in Mcal/d.

    Notes
    -----
    - While this method estimates the release or requirement of NE for uterine weight changes, a note was left in the R code that says it will slightly underestimates the release of NE from the regressing uterus.
    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code: Line 2361.
    - Equation from the Nutrient Requirements of Dairy Cattle book: Equation 20-234 & Table 20-10

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of gestation RE gain
    coeff_dict = {'NE_GrUtWt': 0.95}  
    nd.calculate_Gest_REgain(
        GrUter_BWgain=0.1,  # 0.1 kg/d of uterine body weight gain
        coeff_dict=coeff_dict
    )
    ```
    """
    Gest_REgain = GrUter_BWgain * coeff_dict['NE_GrUtWt']  
    # This will slightly underestimate release of NE from the regressing uterus, Line 2361
    return Gest_REgain


def calculate_Ky_ME_NE(Gest_REgain: float) -> float:
    Ky_ME_NE = 0.14 if Gest_REgain >= 0 else 0.89
    return Ky_ME_NE


def calculate_Gest_MEuse(Gest_REgain: float, Ky_ME_NE: float) -> float:
    """
    *Calculate Gestation (Gest) Metabolizable Energy (ME) Use*

    Calculate the ME requirement for gestation in dairy cows, measured in Megacalories per day (Mcal/d).
    This function determines the ME needed to support gestation, taking into account the reserve energy (RE) gain for gestation and adjusting for energy efficiencies.

    Parameters
    ----------
    Gest_REgain : float
        The reserve energy gain for gestation, in Mcal/d. Normally calculated by 
        [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_Gest_REgain`).

    Returns
    -------
    float
        The ME requirement for gestation, in Mcal/d.

    Notes
    -----
    - The efficiency of energy use (Ky_ME_NE), expressed in Mcal NE/Mcal ME, varies based on whether there is a gain (0.14) or loss (0.89) in gestation energy, reflecting different physiological needs during pregnancy.
    - These efficiencies are derived from Ferrell et al., 1976, with the assumption that the loss efficiency is equivalent to reserve loss efficiency.
    - This method provides insight into the additional ME required to support gestational processes, including uterine growth or regression.
    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code: 
        - Ky_ME_NE = Line 2840
        - Gest_MEuse = Line 2860
    - Equation from the Nutrient Requirements of Dairy Cattle book:
        - Ky_ME_NE = Equation 20-236
        - Gest_MEuse = Equation 20-237

    - TODO: Consider adding the default Ky values to coeff_dict, e.g. Ky_ME_NE_gain = 0.14, Ky_ME_NE_loss = 0.89

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of ME requirement for gestation
    Gest_REgain = 0.2  # 0.2 Mcal/d of RE gain for gestation
    nd.calculate_Gest_MEuse(Gest_REgain)
    ```
    """
    # Gain from Ferrell et al, 1976, and loss assumed = Rsrv loss, Line 2860
    Gest_MEuse = Gest_REgain / Ky_ME_NE
    return Gest_MEuse


def calculate_Trg_Mlk_NEout(
    Trg_MilkProd: float,
    Trg_NEmilk_Milk: float
) -> float:
    """
    *Calculate Target (Trg) Milk Net Energy (NE) Output*

    Calculate the NE requirement for the Target milk production, measured in Megacalories per day (Mcal/d). 
    This function multiplies the target milk production by the target energy output per kg of milk to determine the total energy output required for the specified level of milk production.

    Parameters
    ----------
    Trg_MilkProd : float
        The target milk production, in kg/d.
    Trg_NEmilk_Milk : float
        The target energy output in Mcal per kg of milk. Normally calculated by 
        [](`~nasem_dairy.nasem_equations.milk_equations.calculate_Trg_NEmilk_Milk`).

    Returns
    -------
    float
        The NE requirement for the Target milk production, in Mcal/d.

    Notes
    -----
    - This is based on the Target milk production
    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code: Line 2889.
    - Equation from the Nutrient Requirements of Dairy Cattle book: Equation 20-220

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of NE requirement for milk production
    nd.calculate_Trg_Mlk_NEout(
        Trg_MilkProd=30,  # 30 kg of milk production per day
        Trg_NEmilk_Milk=0.65  # 0.65 Mcal of energy output per kg of milk
    )
    ```
    """
    Trg_Mlk_NEout = Trg_MilkProd * Trg_NEmilk_Milk  # Line 2889
    return Trg_Mlk_NEout


def calculate_Trg_Mlk_MEout(Trg_Mlk_NEout: float, coeff_dict: dict) -> float:
    """
    *Calculate Target (Trg) Milk Metabolizable Energy (ME) Output*

    This function converts NE output required for milk production into its metabolizable energy (ME) equivalent, 
    using a conversion coefficient (Kl_ME_NE) that reflects the efficiency of converting consumed ME into milk NE.

    Parameters
    ----------
    Trg_Mlk_NEout : float
        The NE requirement for milk production, in Mcal/d. Normally calculated by 
        [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_Trg_Mlk_NEout`).
    coeff_dict : dict
        A dictionary containing the conversion coefficient 'Kl_ME_NE' for converting NE to ME for lactation. Defaults to 0.66.

    Returns
    -------
    float
        The ME requirement for milk production, in Mcal/d.

    Notes
    -----
    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code: Line 2890.
    - Reference to specific equations and line in the Nutrient Requirements of Dairy Cattle R Code: 
        - Kl_ME_NE  = Equation 20-223 = 0.66
        - Equation 20-222

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of ME requirement for milk production
    coeff_dict = {'Kl_ME_NE': 0.66}  # Conversion coefficient for NE to ME for lactation
    nd.calculate_Trg_Mlk_MEout(
        Trg_Mlk_NEout=2,  
        coeff_dict=coeff_dict
    )
    ```
    """
    Trg_Mlk_MEout = Trg_Mlk_NEout / coeff_dict['Kl_ME_NE']  
    # Line 2890 Equation 20-222
    return Trg_Mlk_MEout


def calculate_Trg_MEuse(
    An_MEmUse: float, 
    An_MEgain: float,
    Gest_MEuse: float,
    Trg_Mlk_MEout: float
) -> float:
    """
    *Calculate Total Target (Trg) Metabolizable Energy (ME) Use*

    This function sums the ME requirements for maintenance (MEmUse), body and frame gain (MEgain), gestation (MEuse), and target milk output (MEout) to determine the overall ME requirements.

    Parameters
    ----------
    An_MEmUse : float
        The ME required for maintenance, in Mcal/d. Normally calculated by 
        [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_An_MEmUse`).
    An_MEgain : float
        The ME associated with body and frame gain, in Mcal/d. Normally calculated by 
        [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_An_MEgain`).
    Gest_MEuse : float
        The ME required for gestation, in Mcal/d. Normally calculated by 
        [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_Gest_MEuse`).
    Trg_Mlk_MEout : float
        The ME output for milk production, in Mcal/d. Normally calculated by 
        [](`~nasem_dairy.nasem_equations.energy_requirement.calculate_Trg_Mlk_MEout`).

    Returns
    -------
    float
        The total ME requirement, in Mcal/d.

    Notes
    -----
    - Reference to specific line in the Nutrient Requirements of Dairy Cattle R Code: Line 2924.
    - TODO: Check for how to implement with Predicted instead of Target milk?

    Examples
    --------
    ```{python}
    import nasem_dairy as nd

    # Example calculation of total ME requirement
    nd.calculate_Trg_MEuse(
        An_MEmUse=20,     # 20 Mcal/d for maintenance
        An_MEgain=5,      # 5 Mcal/d for body and frame gain
        Gest_MEuse=1,     # 1 Mcal/d for gestation
        Trg_Mlk_MEout=35  # 35 Mcal/d for milk production
    )
    ```
    """
    Trg_MEuse = An_MEmUse + An_MEgain + Gest_MEuse + Trg_Mlk_MEout  # Line 2924
    return Trg_MEuse


def calculate_An_MEmUse_NS(An_NEmUse_NS: float, Km_ME_NE: float) -> float:
    """
    An_MEmUse_NS: Metabolizable energy for maintenance (Mcal/d) NS = non stressed
    """
    An_MEmUse_NS = An_NEmUse_NS / Km_ME_NE  # Line 2846
    return An_MEmUse_NS


def calculate_An_MEmUse_Act(An_NEmUse_Act: float, Km_ME_NE: float) -> float:
    """
    An_MEmUse_Act: Metabolizable energy for activity (Mcal/d)
    """
    An_MEmUse_Act = An_NEmUse_Act / Km_ME_NE  # Line 2847
    return An_MEmUse_Act


def calculate_An_MEmUse_Env(Km_ME_NE: float, coeff_dict: dict) -> float:
    """
    An_MEmUse_Env: Metabolizable energy lost to the environment (Mcal/d)
    
    Examples
    --------
    ```
    coeff_dict = {"An_NEmUse_Env": 0}
    
    calculate_An_MEmUse_Env(
        Km_ME_NE = 0.8, coeff_dict = coeff_dict
    )
    ```
    """
    An_MEmUse_Env = coeff_dict['An_NEmUse_Env'] / Km_ME_NE # Line 2848
    return An_MEmUse_Env


def calculate_An_NEm_ME(An_NEmUse: float, An_MEIn: float) -> float:
    """
    An_NEm_ME: Proportion of MEIn used for maintenance (Mcal/Mcal)
    """
    An_NEm_ME = An_NEmUse / An_MEIn  # proportion of MEIn used for maintenance, Line 2849
    return An_NEm_ME


def calculate_An_NEm_DE(An_NEmUse: float, An_DEIn: float) -> float:
    """
    An_NEm_DE: Proportion of DE intake used for maintenance (Mcal/Mcal)
    """
    An_NEm_DE = An_NEmUse / An_DEIn # proportion of DEIn used for maintenance, Line 2850
    return An_NEm_DE


def calculate_An_NEmNS_DE(An_NEmUse_NS: float, An_DEIn: float) -> float:
    """
    An_NEmNS_DE: NE required for maintenance in an unstressed animal as a proportion of DE intake (Mcal/Mcal)
    """
    An_NEmNS_DE = An_NEmUse_NS / An_DEIn  # Line 2851
    return An_NEmNS_DE


def calculate_An_NEmAct_DE(An_NEmUse_Act: float, An_DEIn: float) -> float:
    """
    An_NEmAct_DE: NE required for maintenance activity as proportion of DE intake (Mcal/Mcal)
    """
    An_NEmAct_DE = An_NEmUse_Act / An_DEIn  # Line 2852
    return An_NEmAct_DE


def calculate_An_NEmEnv_DE(An_DEIn: float, coeff_dict: dict) -> float:
    """
    An_NEmEnv_DE: NE lost to environment as proportion of DE intake (Mcal/Mcal)
    
    Examples
    --------
    ```
    coeff_dict = {"An_NEmUse_Env": 0}
    
    calculate_An_NEmEnv_DE(
        An_DEIn = 50.0, coeff_dict = coeff_dict
    )
    ```
    """
    An_NEmEnv_DE = coeff_dict['An_NEmUse_Env'] / An_DEIn  # Line 2853
    return An_NEmEnv_DE


def calculate_An_NEprod_Avail(An_NEIn: float, An_NEmUse: float) -> float:
    """
    An_NEprod_Avail: NE available for production (Mcal/d)
    """
    An_NEprod_Avail = An_NEIn - An_NEmUse  # Line 2856
    return An_NEprod_Avail


def calculate_An_MEprod_Avail(An_MEIn: float, An_MEmUse: float) -> float:
    """
    An_MEprod_Avail: ME available for production (Mcal/d)
    """
    An_MEprod_Avail = An_MEIn - An_MEmUse  # Line 2857
    return An_MEprod_Avail


def calculate_Gest_NELuse(Gest_MEuse: float, coeff_dict: dict) -> float:
    """
    Gest_NELuse: NE lactation used for gestation (Mcal/d)
    
    Examples
    --------
    ```
    coeff_dict = {"Kl_ME_NE": 0.66}
    
    calculate_Gest_NELuse(
        Gest_MEuse = 100.0, coeff_dict = coeff_dict
    )
    ```
    """
    Gest_NELuse = Gest_MEuse * coeff_dict['Kl_ME_NE']  
    # mcal/d ME required. ??This should not be used, delete. Line 2861
    return Gest_NELuse


def calculate_Gest_NE_ME(Gest_MEuse: float, An_MEIn: float) -> float:
    """
    Gest_NE_ME: Proportion of ME intake used for gestation 
    """
    Gest_NE_ME = Gest_MEuse / An_MEIn 
    # proportion of MEIn used for Gestation, Line 2862
    return Gest_NE_ME


def calculate_Gest_NE_DE(Gest_REgain: float, An_DEIn: float) -> float:
    """
    Gest_NE_DE: Proportion of DE intake retained in gravid uterus tissue (Mcal)
    """
    Gest_NE_DE = Gest_REgain / An_DEIn  
    # proportion of DEIn retained in gravid uterus tissue, Line 2863
    return Gest_NE_DE


def calculate_An_REgain(Body_Fatgain: float, Body_CPgain: float) -> float:
    """
    An_REgain: Retained energy (Mcal/d)
    """
    An_REgain = 9.4 * Body_Fatgain + 5.55 * Body_CPgain  
    # Retained Energy, mcal/d. Does not apply to calves, Line 2866
    return An_REgain


def calculate_Rsrv_NE_DE(Rsrv_NEgain: float, An_DEIn: float) -> float:
    """
    Rsrv_NE_DE: Proportion of DE intake used for reserves gain
    """
    Rsrv_NE_DE = Rsrv_NEgain / An_DEIn  
    # proportion of DEIn used for Reserves gain, Line 2869
    return Rsrv_NE_DE


def calculate_Frm_NE_DE(Frm_NEgain: float, An_DEIn: float) -> float:
    """
    Frm_NE_DE: Proportion of DE intake used for frame gain
    """
    Frm_NE_DE = Frm_NEgain / An_DEIn  
    # proportion of DEIn used for Frame gain, Line 2870
    return Frm_NE_DE


def calculate_Body_NEgain_BWgain(An_REgain: float, Body_Gain: float) -> float:
    """
    Body_NEgain_BWgain: Mcal of NE per kg bodyweight gain 
    """
    if Body_Gain == 0:
        Body_NEgain_BWgain = 0
    else:    
        Body_NEgain_BWgain = An_REgain / Body_Gain  # mcal NE/kg BW gain, Line 2871
    return Body_NEgain_BWgain


def calculate_An_ME_NEg(An_REgain: float, An_MEgain: float) -> float:
    """
    An_ME_NEg: Retained energy as a proportion on ME for gain
    """
    if An_MEgain == 0:
        An_ME_NEg = 0
    else:    
        An_ME_NEg = An_REgain / An_MEgain  
    # A weighted average efficiency based on user entered or prediction frame
    # and reserves gains. Cows only at this time, Line 2875
    return An_ME_NEg


def calculate_Rsrv_NELgain(Rsrv_MEgain: float, coeff_dict: dict) -> float:
    """
    Rsrv_NELgain: NEL required for body reserve gain (Mcal/d)

    Examples
    --------
    ```
    coeff_dict = {"Kl_ME_NE": 0.66}
    
    calculate_Rsrv_NELgain(
        Rsrv_MEgain = 120.0, coeff_dict = coeff_dict
    )
    ```
    """
    Rsrv_NELgain = Rsrv_MEgain * coeff_dict['Kl_ME_NE']  
    # express body energy required in NEL terms, Line 2877
    return Rsrv_NELgain


def calculate_Frm_NELgain(Frm_MEgain: float, coeff_dict: dict) -> float:
    """
    Frm_NELgain: NEL required for frame gain (Mcal/d)

    Examples
    --------
    ```
    coeff_dict = {"Kl_ME_NE": 0.66}
    
    calculate_Frm_NELgain(
        Frm_MEgain = 80.0, coeff_dict = coeff_dict
    )
    ```
    """
    Frm_NELgain = Frm_MEgain * coeff_dict['Kl_ME_NE']  # Line 2878
    return Frm_NELgain


def calculate_An_NELgain(An_MEgain: float, coeff_dict: dict) -> float:
    """
    An_NELgain: NEL required for bodyweight gain (Mcal/d)

    Examples
    --------
    ```
    coeff_dict = {"Kl_ME_NE": 0.66}
    
    calculate_An_NELgain(
        An_MEgain = 150.0, coeff_dict = coeff_dict
    )
    ```
    """
    An_NELgain = An_MEgain * coeff_dict['Kl_ME_NE']  # Line 2879
    return An_NELgain


def calculate_An_NEgain_DE(An_REgain: float, An_DEIn: float) -> float:
    """
    An_NEgain_DE: Retained energy as a proportion of DE intake
    """
    An_NEgain_DE = An_REgain / An_DEIn  # Line 2881
    return An_NEgain_DE


def calculate_An_NEgain_ME(An_REgain: float, An_MEIn: float) -> float:
    """
    An_NEgain_ME: Retained energy for gain as a proportion of ME intake
    """
    An_NEgain_ME = An_REgain / An_MEIn  # Line 2882
    return An_NEgain_ME


def calculate_An_MEuse(
    An_MEmUse: float, 
    An_MEgain: float, 
    Gest_MEuse: float,
    Mlk_MEout: float
) -> float:
    """
    An_MEuse: Total ME use (Mcal/d)
    """
    An_MEuse = An_MEmUse + An_MEgain + Gest_MEuse + Mlk_MEout  # Line 2923
    return An_MEuse


def calculate_An_NEuse(
    An_NEmUse: float, 
    An_REgain: float, 
    Gest_REgain: float,
    Mlk_NEout: float
) -> float:
    """
    An_NEuse: Total NE use (Mcal/d)
    """
    An_NEuse = An_NEmUse + An_REgain + Gest_REgain + Mlk_NEout  # Line 2925
    return An_NEuse


def calculate_Trg_NEuse(
    An_NEmUse: float, 
    An_REgain: float, 
    Gest_REgain: float,
    Trg_Mlk_NEout: float
) -> float:
    """
    Trg_NEuse: Target total NE use (Mcal/d)
    """
    Trg_NEuse = An_NEmUse + An_REgain + Gest_REgain + Trg_Mlk_NEout  # Line 2926
    return Trg_NEuse


def calculate_An_NELuse(An_MEuse: float, coeff_dict: dict) -> float:
    """
    An_NELuse: Total NEL requirement (Mcal/d)
    
    Examples
    --------
    ```
    coeff_dict = {"Kl_ME_NE": 0.66}
    
    calculate_An_NELuse(
        An_MEuse = 100.0, coeff_dict = coeff_dict
    )
    ```
    """
    An_NELuse = An_MEuse * coeff_dict['Kl_ME_NE']  # Line 2927
    return An_NELuse


def calculate_Trg_NELuse(Trg_MEuse: float, coeff_dict: dict) -> float:
    """
    Trg_NELuse: Target NEL requirement (Mcal/d)

    Examples
    --------
    ```
    coeff_dict = {"Kl_ME_NE": 0.66}
    
    calculate_Trg_NELuse(
        Trg_MEuse = 90.0, coeff_dict = coeff_dict
    )
    ```
    """
    Trg_NELuse = Trg_MEuse * coeff_dict['Kl_ME_NE']  # Line 2928
    return Trg_NELuse


def calculate_An_NEprod_GE(
    An_NEuse: float, 
    An_NEmUse: float,
    An_GEIn: float
) -> float:
    """
    An_NEprod_GE: Total efficiency of GE use (predicted NE use)
    """
    An_NEprod_GE = (An_NEuse - An_NEmUse) / An_GEIn  
    # Total efficiency of GE use, Line 2930
    return An_NEprod_GE


def calculate_Trg_NEprod_GE(
    Trg_NEuse: float, 
    An_NEmUse: float,
    An_GEIn: float
) -> float:
    """
    Trg_NEprod_GE: Total efficiency of GE use (target NE use)
    """
    Trg_NEprod_GE = (Trg_NEuse - An_NEmUse) / An_GEIn  
    # Total efficiency of GE use, Line 2931
    return Trg_NEprod_GE


def calculate_An_NEmlk_GE(Mlk_NEout: float, An_GEIn: float) -> float:
    """
    An_NEmlk_GE: Efficiency of GE use for predicted milk NE
    """
    An_NEmlk_GE = Mlk_NEout / An_GEIn  
    # Efficiency of GE use for milk NE, Line 2932
    return An_NEmlk_GE


def calculate_Trg_NEmlk_GE(Trg_Mlk_NEout: float, An_GEIn: float) -> float:
    """
    Trg_NEmlk_GE: Efficiency of GE use for target milk NE
    """
    Trg_NEmlk_GE = Trg_Mlk_NEout / An_GEIn  
    # Efficiency of GE use for milk NE, Line 2933
    return Trg_NEmlk_GE


def calculate_An_MEbal(An_MEIn: float, An_MEuse: float) -> float:
    """
    An_MEbal: ME balance (Mcal/d)
    """
    An_MEbal = An_MEIn - An_MEuse  # Line 2935
    return An_MEbal


def calculate_An_NELbal(An_MEbal: float, coeff_dict: dict) -> float:
    """
    An_NELbal: NEL balance (Mcal/d)
    
    Examples
    --------
    ```
    coeff_dict = {"Kl_ME_NE": 0.66}
    
    calculate_An_NELbal(
        An_MEbal = 50.0, coeff_dict = coeff_dict
    )
    ```
    """
    An_NELbal = An_MEbal * coeff_dict['Kl_ME_NE']  # Line 2936
    return An_NELbal


def calculate_An_NEbal(An_NEIn: float, An_NEuse: float) -> float:
    """
    An_NEbal: NE balance (Mcal/d)
    """
    An_NEbal = An_NEIn - An_NEuse  # Line 2937
    return An_NEbal


def calculate_Trg_MEbal(An_MEIn: float, Trg_MEuse: float) -> float:
    """
    Trg_MEbal: Target ME balance (Mcal/d)
    """
    Trg_MEbal = An_MEIn - Trg_MEuse  # Line 2939
    return Trg_MEbal


def calculate_Trg_NELbal(Trg_MEbal: float, coeff_dict: dict) -> float:
    """
    Trg_NELbal: Target NEL balance (Mcal/d)
    
    Examples
    --------
    ```
    coeff_dict = {"Kl_ME_NE": 0.66}
    
    calculate_Trg_NELbal(
        Trg_MEbal = 70.0, coeff_dict = coeff_dict
    )
    ```
    """
    Trg_NELbal = Trg_MEbal * coeff_dict['Kl_ME_NE']  # Line 2940
    return Trg_NELbal


def calculate_Trg_NEbal(An_NEIn: float, Trg_NEuse: float) -> float:
    """
    Trg_NEbal: Target NE balance (Mcal/d)
    """
    Trg_NEbal = An_NEIn - Trg_NEuse  # Line 2941
    return Trg_NEbal


def calculate_An_MPuse_MEuse(An_MPuse_g: float, An_MEuse: float) -> float:
    """
    An_MPuse_MEuse: Ratio of MP use to ME use (g/Mcal)
    """
    An_MPuse_MEuse = An_MPuse_g / An_MEuse  # g/mcal, Line 2943
    return An_MPuse_MEuse


def calculate_Trg_MPuse_MEuse(An_MPuse_g_Trg: float, An_MEuse: float) -> float:
    """
    Trg_MPuse_MEuse: Target ratio of MP use to ME use (g/Mcal)
    """
    Trg_MPuse_MEuse = An_MPuse_g_Trg / An_MEuse  # Line 2944
    return Trg_MPuse_MEuse


def calculate_Km_ME_NE_Clf(
    An_ME_ClfDry: float, 
    An_NE_ClfDry: float, 
    Dt_DMIn_ClfLiq: float, 
    Dt_DMIn_ClfStrt: float
) -> float:
    """Conversion efficiency of ME to NE for maintenance, calves"""
    if (An_ME_ClfDry > 0 and 
        An_NE_ClfDry > 0 and 
        Dt_DMIn_ClfLiq == 0
        ):
        Km_ME_NE_Clf = An_NE_ClfDry / An_ME_ClfDry
    elif Dt_DMIn_ClfStrt == 0 and Dt_DMIn_ClfLiq > 0:
        Km_ME_NE_Clf = 0.723
    else:
        Km_ME_NE_Clf = 0.69 
    return Km_ME_NE_Clf


def calculate_Km_ME_NE(An_StatePhys: str) -> float:
    """Conversion efficiency of ME to NE for maintenance"""
    if An_StatePhys == "Lactating Cow" or An_StatePhys == "Dry Cow":
        Km_ME_NE = 0.66 # Km_ME_NE_Cow
    else:
        Km_ME_NE = 0.63 # Km_ME_NE_Heif
    return Km_ME_NE


def calculate_Trg_NEmilkOut(
    Trg_NEmilk_Milk: float, 
    Trg_MilkProd: float
) -> float:
    Trg_NEmilkOut = Trg_NEmilk_Milk * Trg_MilkProd  # Line 387
    return Trg_NEmilkOut



"""Directed Acyclic Graph (DAG) management for the NASEM model.

This module provides functionality for creating, managing, and visualizing a 
Directed Acyclic Graph (DAG) that represents the dependencies and execution 
order of calculations within the NASEM model. It leverages the `graph-tool` 
library for efficient graph manipulation and visualization.

Classes:
    ModelDAG: Manages the creation and validation of a DAG for the NASEM model, 
              including methods for generating calculation orders and 
              dynamically creating functions.
"""

import ast
import glob
import inspect
import os
from typing import Dict, List, Any, Tuple, Optional, Callable

import pandas as pd

import nasem_dairy as nd
import nasem_dairy.model.input_definitions as expected
import nasem_dairy.model_output.ModelOutput as output

try:
    import graph_tool.all as graph_tool
except ImportError: # pragma: no cover
    raise ImportError(
        "The 'graph-tool' package is required for this module. "
        "Install it with `poetry install --extras dag` or "
        "`pip install nasem-dairy[dag]`."
    )
# NOTE: Since ModelDAG is a subpackage we should inlcude the following code in any 
# modules where it is used. This will ensure a meaningful error is raised in cases
# where graph-tool is not imported (nd.ModelDAG is set to None in __init__.py) but
# a user tries to run code that requires nd.ModelDAG.
# if nd.ModelDAG is None:
#     raise ImportError(
#         "ModelDAG requires the 'graph-tool' package. Please install it with "
#         "`poetry install --extras dag`."
#     )

module_colour_map = {
            "amino_acid": [1.0, 0.0, 0.0, 0.7],            # Bright Red
            "animal": [0.0, 0.5, 1.0, 0.7],                # Sky Blue
            "body_composition": [0.0, 1.0, 0.0, 0.7],      # Bright Green
            "coefficient_adjustment": [0.6, 0.2, 0.8, 0.7],# Lavender
            "dry_matter_intake": [1.0, 0.5, 0.0, 0.7],     # Orange
            "energy_requirement": [0.2, 0.7, 0.2, 0.7],    # Forest Green
            "fecal": [0.8, 0.4, 0.0, 0.7],                 # Burnt Orange
            "gestation": [1.0, 0.2, 0.6, 0.7],             # Pink
            "infusion": [0.4, 0.0, 0.8, 0.7],              # Deep Purple
            "manure": [0.55, 0.27, 0.07, 0.7],             # Brown
            "methane": [0.2, 0.8, 0.8, 0.7],               # Aqua
            "microbial_protein": [0.6, 0.6, 0.2, 0.7],     # Olive Green
            "micronutrient_requirement": [0.8, 0.2, 0.0, 0.7], # Rust
            "milk": [0.5, 1.0, 0.5, 0.7],                  # Light Green
            "nutrient_intakes": [0.0, 0.4, 0.8, 0.7],      # Royal Blue
            "protein_requirement": [1.0, 0.7, 0.0, 0.7],   # Amber
            "protein": [0.8, 0.0, 0.8, 0.7],               # Magenta
            "report": [0.3, 0.3, 0.3, 0.7],                # Charcoal Gray
            "rumen": [0.2, 0.8, 1.0, 0.7],                 # Cyan
            "urine": [0.4, 0.8, 0.4, 0.7],                 # Sage Green
            "water": [0.0, 0.5, 1.0, 0.7],                 # Deep Sky Blue
            "Constants": [0.6, 0.6, 0.6, 1.0],             # Light Gray
            "Inputs": [0.2, 0.2, 0.2, 1.0]                 # Dark Gray
            }

class ModelDAG:
    """
    Manages the creation, validation, and visualization of a Directed Acyclic Graph (DAG) 
    for the NASEM model.

    The `ModelDAG` class is responsible for constructing a DAG that represents the 
    dependencies and execution order of calculations within the NASEM model. It uses 
    the `graph-tool` library for efficient graph manipulation and visualization, and 
    provides various methods to validate the structure, generate calculation orders, 
    and create dynamic functions based on the DAG.

    Attributes:
        aa_list (List[str]): A list of essential amino acids used in various calculations.
        module_colour_map (dict): A dictionary mapping module names to RGBA color values 
            for graph visualization.
        possible_user_inputs (Dict[str, Dict[str, Any]]): A dictionary of possible user 
            input structures, mapping names to their respective schema.
        possible_constants (Dict[str, Dict[str, Any]]): A dictionary of possible constants, 
            mapping names to their respective schema.
        user_inputs (List[str]): A list of user input keys extracted from possible user input structures.
        modules (List[str]): A list of Python file paths in the specified directory, excluding `__init__.py`.
        dag_data (pd.DataFrame): A DataFrame containing parsed data for each variable in the DAG.
        dag (graph_tool.Graph): The Directed Acyclic Graph representing the dependencies of variables.

    Methods:
        __init__(self, path: str = "./src/nasem_dairy/nasem_equations", colour_map: dict = module_colour_map):
            Initializes the `ModelDAG` instance by collecting data and creating the DAG.
        
        _get_variable_names(self) -> List[str]:
            Retrieves the variable names needed to build the DAG.
        
        _get_py_files(self, path: str) -> List[str]:
            Retrieves a list of Python files from the specified directory.
        
        _get_dict_keys(self, node: ast.AST, id_to_check: str, check_fstring: bool = True) -> List[str]:
            Extracts keys from a dictionary accessed in an AST node.
        
        _generate_user_inputs_list(self, possible_user_inputs: Dict[str, Any]) -> List[str]:
            Generates a list of user input keys from possible user input structures.
        
        _create_function_entry(self, node: ast.FunctionDef) -> Tuple[Optional[str], List[str], List[str], List[str]]:
            Collects required data for regular functions from an AST node.
        
        _parse_nasem_equations(self, py_files: List[str], variables: pd.DataFrame) -> pd.DataFrame:
            Parses NASEM equations from Python files and extracts data for the DAG.
        
        _create_dag(self, data: pd.DataFrame) -> graph_tool.Graph:
            Creates a Directed Acyclic Graph (DAG) from the provided data.
        
        draw_dag(self, output_path: str) -> None:
            Draws and saves the DAG to the specified output path.
        
        validate_dag(self) -> None:
            Validates the DAG structure by checking for cycles, verifying the topological order, and ensuring connectivity.
        
        get_calculation_order(self, target_variable: str, report: bool = True) -> Dict[List[str], Dict[str, Dict[str, Any]]]:
            Determines the calculation order for a given target variable using depth-first search on the DAG.
        
        create_function(self, target_variable: str) -> Callable[..., Any]:
            Creates a dynamically generated function to calculate the target variable based on the DAG structure.
    """
    ### Initalization ###
    def __init__(
        self, 
        path:str = "./src/nasem_dairy/nasem_equations", 
        colour_map: dict = module_colour_map
    ):
        """
        Collect data for DAG and create graph.
        """
        self.aa_list = [
            "Arg", "His", "Ile", "Leu", "Lys", "Met", "Phe", "Thr", "Trp", "Val"
            ]
        self.module_colour_map = colour_map
        self.possible_user_inputs = {
            "animal_input": expected.AnimalInput.__annotations__.copy(),
            "equation_selection": (
                expected.EquationSelection.__annotations__.copy()
                ),
            "infusion_input": expected.InfusionInput.__annotations__.copy(),
            "user_diet": expected.UserDietSchema.copy(),
            "feed_library": expected.FeedLibrarySchema.copy()
        }
        self.possible_constants = {
            "coeff_dict": expected.CoeffDict.__annotations__.copy(),
            "infusion_dict": expected.InfusionDict.__annotations__.copy(),
            "mpnp_efficiency_dict": (
                expected.MPNPEfficiencyDict.__annotations__.copy()
                ),
            "mprt_coeff_dict": expected.mPrtCoeffDict.__annotations__.copy(),
            "f_imb": expected.f_Imb.copy()
        }
        self.user_inputs = self._generate_user_inputs_list(
            self.possible_user_inputs
            )

        variable_names = self._get_variable_names()
        variables = pd.DataFrame(variable_names, columns=["Name"])

        # Collect data for DAG
        self.modules = self._get_py_files(path)  
        self.dag_data = self._parse_nasem_equations(self.modules, variables)
        self.dag_data = self.dag_data.dropna(axis=0)
        self.dag = self._create_dag(self.dag_data)

    def _get_variable_names(self) -> List[str]:
        """
        Retrieve the variable names needed to build the DAG.

        Returns:
            A list of variable names required for the DAG
        """
        user_diet, animal_input, eqn_selection, inf_input = nd.demo(
            "lactating_cow_test"
            )
        output = nd.nasem(
            user_diet = user_diet, 
            animal_input = animal_input, 
            equation_selection = eqn_selection, 
            coeff_dict = nd.coeff_dict,
            infusion_input=inf_input
            )
        variables = output.export_variable_names()
        # These variables are necessary for the DAG but not directly available
        additional_vars = [
            "Abs_EAA2_HILKM_g" ,"Abs_EAA2_RHILKM_g", "Abs_EAA2_HILKMT_g", 
            "An_GasEOut_Dry", "An_GasEOut_Lact", "An_GasEOut_Heif"
            ]
        variables.extend(additional_vars)
        return variables

    def _get_py_files(self, path: str) -> List[str]:
        """
        Retrieve a list of Python files from the specified directory.

        Args:
            path: A string representing the directory path to search for Python files.

        Returns:
            A list of Python file paths in the specified directory, excluding `__init__.py`.
        """
        py_files = glob.glob(os.path.join(path, "*.py"))
        return [
            file for file in py_files 
            if os.path.basename(file) != "__init__.py"
            ]

    def _get_dict_keys(
        self, 
        node: ast.AST, 
        id_to_check: str, 
        check_fstring: bool = True
    ) -> List[str]:
        """
        Extract keys from a dictionary accessed in an AST node.

        Args:
            node: The AST node to analyze for dictionary key accesses.
            id_to_check: The identifier of the dictionary to check.
            check_fstring: Whether to check for keys formatted with f-strings.
        
        Returns:
            A list of keys accessed from the specified dictionary.
        """
        coeff_keys = []
        for body_item in node.body:
            for sub_node in ast.walk(body_item):
                if isinstance(sub_node, ast.Subscript):
                    if (isinstance(sub_node.value, ast.Name) and
                        sub_node.value.id == id_to_check):
                        
                        # Get key when key is a string: dictionary["key"]
                        if (isinstance(sub_node.slice, ast.Constant) and 
                            isinstance(sub_node.slice.value, str)):
                            coeff_keys.append(sub_node.slice.value)

                        elif check_fstring:
                            # Create keys when key is an f-string: dictionary[f"text{aa}text]
                            if isinstance(sub_node.slice, ast.JoinedStr):
                                f_string = sub_node.slice.values
                                for section in f_string:

                                    # If ast.FormattedValue == "aa" assume iterating self.aa_list
                                    if (isinstance(section, ast.FormattedValue) and 
                                        section.value.id == "aa"):
                                        for aa in self.aa_list:
                                            formatted_coeff = ""
                                            for val in f_string:
                                                if (isinstance(val, ast.Constant) and 
                                                    isinstance(val.value, str)):
                                                    formatted_coeff += val.value

                                                elif (isinstance(val, ast.FormattedValue) and 
                                                    val.value.id == "aa"):
                                                    formatted_coeff += aa

                                            coeff_keys.append(formatted_coeff)
        return coeff_keys

    def _generate_user_inputs_list(
        self, 
        possible_user_inputs: Dict[str, Any]
    ) -> List[str]:
        """
        Generate a list of user input keys from possible user input structures.

        This method iterates over the possible user input structures from the
        input definitions module to extract the keys representing user inputs.

        Args:
            possible_user_inputs: A dictionary mapping structure names to their
                corresponding input structures, which can be either dictionaries
                or objects with `__annotations__`.

        Returns:
            A list of all user input keys found in the input structures.
        """
        user_inputs = []
        for structure_name, structure in possible_user_inputs.items():
            if isinstance(structure, dict):
                user_inputs.extend(structure.keys())
        return user_inputs

    def _create_function_entry(
        self, 
        node: ast.FunctionDef
    ) -> Tuple[Optional[str], List[str], List[str], List[str]]:
        """
        Collect required data for regular functions from an AST node.

        This method processes an AST node representing a function definition
        to extract its return variable, arguments, constants, and inputs.

        Args:
            node: An AST node representing a function definition.

        Returns:
            A tuple containing:
                - The return variable name (or None if not found).
                - A list of arguments that are not constants or user inputs.
                - A list of keys related to constants required by the function.
                - A list of arguments identified as user inputs.
        """
        args = [arg.arg for arg in node.args.args if arg.arg not in ["aa_list"]]

        # Extract names of constants from function definition
        coeff_keys = []
        dicts_to_check = [
            "coeff_dict", "mPrt_coeff", "MP_NP_efficiency_dict"
            ]
        for dictionary in dicts_to_check:
            if dictionary in args:
                keys = self._get_dict_keys(node, dictionary)
                coeff_keys.extend(keys)
                args.remove(dictionary)

        # Check for f_Imb seperatly as not a dictionary
        constants_series = ["f_Imb", "SIDig_values"]
        for constant in constants_series:
            if constant in args:
                coeff_keys.append(constant)
                args.remove(constant)

        # Extract names of keys when dict is passed as arg
        dicts_to_check = ["infusion_data", "diet_data", "feed_data", "an_data"]
        for dictionary in dicts_to_check:
            if dictionary in args:
                keys = self._get_dict_keys(node, dictionary)
                args.extend(keys)
                args.remove(dictionary)

        # Check if any args are model inputs
        inputs = [arg for arg in args if arg in self.user_inputs]
        args = [arg for arg in args if arg not in self.user_inputs]

        # Get the return value
        return_var = None
        for body_item in node.body:
            if isinstance(body_item, ast.Return):
                if isinstance(body_item.value, ast.Name):
                    return_var = body_item.value.id

        return return_var, args ,coeff_keys, inputs

    def _parse_nasem_equations(
        self, 
        py_files: List[str], 
        variables: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Parse NASEM equations from Python files and extract data for the DAG.

        This function parses through Python files in the NASEM equations directory 
        to retrieve data used for plotting the DAG. It returns a DataFrame with 
        the relevant information for each variable listed in the variables DataFrame.

        Args:
            py_files: A list of Python file paths to parse.
            variables: A DataFrame containing the variables to include in the DAG.

        Returns:
            A DataFrame containing the parsed data, including the module name, 
            function name, arguments, constants, and inputs associated with each 
            variable.
        """
        dag_data = variables.copy()
        dag_data["Module"] = None
        dag_data["Function"] = None
        dag_data["Arguments"] = None
        dag_data["Constants"] = None
        dag_data["Inputs"] = None

        for py_file in py_files:
            with open(py_file, "r") as file:
                tree = ast.parse(file.read())
                module_name = py_file.split("/")[-1].replace(".py", "")

            for node in ast.walk(tree):
                if isinstance(node, ast.FunctionDef):
                    function_name = node.name
                    return_var, args ,coeff_keys, inputs = (
                        self._create_function_entry(node)
                        )
                    if return_var and return_var in dag_data["Name"].values:
                        idx = dag_data[dag_data["Name"] == return_var].index[0]
                        dag_data.at[idx, "Module"] = module_name
                        dag_data.at[idx, "Function"] = function_name
                        dag_data.at[idx, "Arguments"] = (
                            list(set(args)) if args else args
                            )
                        dag_data.at[idx, "Constants"] = (
                            list(set(coeff_keys)) if coeff_keys else coeff_keys
                            )
                        dag_data.at[idx, "Inputs"] = (
                            list(set(inputs)) if inputs else inputs
                            )
                                       
        return dag_data

    def _create_dag(self, data: pd.DataFrame) -> graph_tool.Graph:
        """
        Create a Directed Acyclic Graph (DAG) from the provided data.

        This function constructs a DAG using the variable names, functions, 
        modules, constants, and inputs from the provided data. Each unique 
        variable name, constant, and input is added as a vertex in the graph, 
        and edges are added based on the dependencies specified in the 
        'Arguments', 'Constants', and 'Inputs' columns.

        Args:
            data: A DataFrame containing the variables, functions, modules, 
                constants, and inputs for the DAG.

        Returns:
            A graph-tool Graph object representing the DAG.
        """
        dag = graph_tool.Graph(directed=True)

        # Create a dictionary to map variable names to graph vertices
        name_to_vertex = {}
        vertex_labels = dag.new_vertex_property("string")
        vertex_functions = dag.new_vertex_property("string")
        vertex_module = dag.new_vertex_property("string")
        vertex_colors = dag.new_vertex_property("vector<double>")

        # Add vertices for each unique variable name in the Name column
        for index, row in data.iterrows():
            name = row['Name']
            function = row["Function"]
            module = row['Module']
            vertex = dag.add_vertex()
            name_to_vertex[name] = vertex
            vertex_labels[vertex] = name
            vertex_functions[vertex] = function
            vertex_module[vertex] = module
            vertex_colors[vertex] = self.module_colour_map.get(
                module, [0, 0, 0, 0]
                )

        # Add vertices for Constants and Inputs
        for column in ["Constants", "Inputs"]:
            for values in data[column]:
                for value in values:
                    if value not in name_to_vertex:
                        vertex = dag.add_vertex()
                        name_to_vertex[value] = vertex
                        vertex_labels[vertex] = value
                        vertex_functions[vertex] = ""
                        vertex_module[vertex] = column
                        vertex_colors[vertex] = self.module_colour_map.get(
                            column, [0.5, 0.5, 0.5, 0.5]
                            )

        # Add edges based on the Arguments column
        for index, row in data.iterrows():
            src_vertex = name_to_vertex[row['Name']]
            arguments = row['Arguments']
            constants = row["Constants"]
            inputs = row["Inputs"]

            for arg in arguments:
                if arg in name_to_vertex:
                    dst_vertex = name_to_vertex[arg]
                    dag.add_edge(dst_vertex, src_vertex)

            for constant in constants:
                if constant in name_to_vertex:
                    dst_vertex = name_to_vertex[constant]
                    dag.add_edge(dst_vertex, src_vertex)

            for input_val in inputs:
                if input_val in name_to_vertex:
                    dst_vertex = name_to_vertex[input_val]
                    dag.add_edge(dst_vertex, src_vertex)

            self.name_to_vertex = name_to_vertex 
            self.vertex_labels = vertex_labels 
            self.vertex_functions = vertex_functions
            self.vertex_module = vertex_module
            self.vertex_colors = vertex_colors

        return dag
        
    ### Visualization ###
    def draw_dag(self, output_path: str) -> None:
        """
        Draw and save the DAG to the specified output path.

        This method generates a visual representation of the DAG using the 
        `graph_tool` library, with custom styling for vertices and layout. 
        The graph is saved as an image file at the specified output path.

        Args:
            output_path: The file path where the DAG image will be saved.
        """
        pos = graph_tool.sfdp_layout(self.dag)
        graph_tool.graph_draw(self.dag, 
                              pos=pos,
                              vertex_text=self.vertex_labels, 
                              vertex_font_size=12, 
                              vertex_size=10,
                              vertex_fill_color=self.vertex_colors,
                              output_size=(8000, 8000), 
                              bg_color=[0.9, 0.9, 0.9, 1],
                              output=output_path
                              )
        print(
            "Graph has", self.dag.num_vertices(), 
            "vertices and", self.dag.num_edges(), "edges."
            )
        print(f"Graph image saved to {output_path}")

    ### Validation ###
    def validate_dag(self) -> None:
        """
        Validate the DAG structure.

        This method performs several checks to ensure the integrity of the DAG:
        - Checks for cycles in the DAG.
        - Validates the topological sort of the DAG.
        - Ensures that all nodes and edges are correctly connected.
        """
        self._check_for_cycles()
        self._validate_topological_order()
        self._verify_connectivity()

    def _check_for_cycles(self) -> None:
        """
        Check for cycles in the DAG.

        This method checks whether the DAG contains any cycles. If a cycle is 
        detected, it raises a ValueError with details of the cycle(s) found.

        Raises:
            ValueError: If one or more cycles are detected in the DAG.
        """
        if graph_tool.is_DAG(self.dag):
            print("No cycles detected.")
        else:
            cycles = list(graph_tool.all_circuits(self.dag))
            if cycles:
                cycle_strs = []
                for cycle in cycles:
                    cycle_vertices = [
                        self.vertex_labels[vertex] for vertex in cycle
                        ]
                    cycle_strs.append(" -> ".join(cycle_vertices))
                raise ValueError(
                    f"Cycles detected in the DAG:\n" + "\n".join(cycle_strs)
                    )

    def _validate_topological_order(self) -> None:
        """
        Validate the topological order of the DAG.

        This method attempts a topological sort of the DAG. If the sort fails, 
        it raises a ValueError, which may indicate the presence of cycles or 
        other inconsistencies in the DAG structure.

        Raises:
            ValueError: If the topological sort fails
        """
        try:
            order = graph_tool.topological_sort(self.dag)
        except ValueError as e:
            raise ValueError(
                "Topological sort failed. DAG may contain cycles or other "
                "inconsistencies."
                ) from e
    
    def _verify_connectivity(self) -> None:
        """
        Ensure all vertices are correctly connected according to the DAG data.

        This method verifies the connectivity of the DAG by checking that:
        - All vertices are connected according to the expected inputs.
        - There are no isolated vertices (vertices with no edges).
        - Each vertex has the correct number of incoming and outgoing edges.

        Raises:
            ValueError: If any vertex is isolated or has an incorrect number 
            of incoming or outgoing edges.
        """
        vertex_data = {
            row["Name"]: {
                "expected_inputs": (
                    row["Arguments"] + row["Constants"] + row["Inputs"]
                    ),
                "outgoing_count": 0
                }
                for index, row in self.dag_data.iterrows()
            }
        # Populate outgoing_count based on references in other vertices' expected_inputs
        for name, data in vertex_data.items():
            for reference in data["expected_inputs"]:
                if reference in vertex_data:
                    vertex_data[reference]["outgoing_count"] += 1
       
        # Check for isolated vertices
        for name, vertex in self.name_to_vertex.items():
            if vertex.out_degree() == 0 and vertex.in_degree() == 0:
                raise ValueError(f"Vertex {name} is isolated (no edges).")

        # Check each vertex has expected number of incoming edges
        for name, data in vertex_data.items():
            if name in self.name_to_vertex:
                vertex = self.name_to_vertex[name]
                expected_incoming_edges = len(data["expected_inputs"])
                actual_incoming_edges = vertex.in_degree()

                if actual_incoming_edges != expected_incoming_edges:
                    actual_incoming_names = [
                        self.vertex_labels[edge.source()] 
                        for edge in vertex.in_edges()
                    ]
                    missing_edges = [
                        edge for edge in data["expected_inputs"] 
                        if edge not in actual_incoming_names
                        ]

                    if missing_edges:
                        print(f"Missing expected edges from: {missing_edges}")

                    raise ValueError(
                        f"Vertex {name} has {actual_incoming_edges} incoming edges, "
                        f"but {expected_incoming_edges} were expected."
                    )
            else:
                print(f"{name} was not found in self.name_to_vertex")

        # Check each vertex has the expected number of outgoing edges
        for name, data in vertex_data.items():
            if name in self.name_to_vertex:
                vertex = self.name_to_vertex[name]
                expected_outgoing_edges = data["outgoing_count"]
                actual_outgoing_edges = vertex.out_degree()

                if actual_outgoing_edges != expected_outgoing_edges:
                    print(
                        f"Vertex {name} has an incorrect number of outgoing edges."
                        )
                    print(f"Expected {expected_outgoing_edges} outgoing edges.")
                    print(f"Actual {actual_outgoing_edges} outgoing edges.")

                    raise ValueError(
                        f"Vertex {name} has {actual_outgoing_edges} outgoing edges, "
                        f"but {expected_outgoing_edges} were expected."
                    )

        print("Connectivity verification completed.")

    ### Tools ###
    def get_calculation_order(
        self, 
        target_variable: str, 
        report: bool = True
    ) -> Dict[List[str], Dict[str, Dict[str, Any]]]:
        """
        Determine the calculation order for a given target variable.

        This method performs a depth-first search (DFS) on the DAG to determine
        the order of function calls and the required inputs and constants needed
        to calculate the specified target variable.

        Args:
            target_variable: The name of the target variable to calculate.
            report: Whether to print a report of the calculation order. 
                Default is True.

        Returns:
            dict:
                - 'functions_order': A list of functions in the order they should 
                be called.
                - 'user_inputs': A dictionary of required user input structures 
                and their respective fields.
                - 'constants': A dictionary of required constants and their 
                respective fields.
        """
        def depth_first_search(vertex: graph_tool.Vertex) -> None: # type:ignore
            """
            Perform a depth-first search to determine dependencies.

            This helper function traverses the graph to find all dependencies for
            the given vertex, recording user inputs, constants, and functions 
            necessary to calculate the target variable.

            Args:
                vertex: The current vertex in the graph being explored.
            """
            if vertex in visited:
                return
            visited.add(vertex)

            # Traverse all incoming edges to find dependencies
            for edge in vertex.in_edges():
                source_vertex = edge.source()
                depth_first_search(source_vertex)

            vertex_name = self.vertex_labels[vertex]
            vertex_function = self.vertex_functions[vertex]
            vertex_module = self.vertex_module[vertex]

            if vertex_module == "Inputs":
                user_inputs.add(vertex_name)
            elif vertex_module == "Constants":
                constants.add(vertex_name)
            elif vertex_function: 
                functions_order.append(vertex_function)


        def print_report(
            target_variable: str, 
            functions_order: List[str], 
            sorted_user_inputs: Dict[str, Dict[str, Optional[str]]], 
            sorted_constants: Dict[str, Optional[Dict[str, Optional[str]]]]
        ) -> None:
            """
            Print a formatted report of the calculation requirements.

            This function generates a report detailing the order of functions to 
            be called, as well as the required user inputs and constants for 
            calculating the specified target variable.

            Args:
                target_variable: The name of the target variable.
                functions_order: A list of functions in the order they should be called.
                sorted_user_inputs: A dictionary of required user input structures 
                    and their respective fields.
                sorted_constants: A dictionary of required constants and their 
                    respective fields or special structures.
            """
            print(f"\nRequirements for Calculating {target_variable}")
            print("\nOrder of Functions to Call:")
            for i, function in enumerate(functions_order, 1):
                print(f"  {i}. {function}")
            
            if sorted_user_inputs:
                print("\nRequired User Inputs:")
                for key in sorted_user_inputs:
                    print(f"  {key}:")
                    for field in sorted_user_inputs[key]:
                        print(f"    - {field}")
            
            if sorted_constants:
                print("\nRequired Constants:")
                for key in sorted_constants:
                    print(f"  {key}:")
                    if isinstance(sorted_constants[key], dict):
                        for field in sorted_constants[key]:
                            print(f"    - {field}")
                    else:
                        print(f"    - {key}")


        if target_variable not in self.name_to_vertex:
            raise ValueError(
                f"Variable '{target_variable}' not found in the DAG."
                )

        target_vertex = self.name_to_vertex[target_variable]
        functions_order = []
        user_inputs = set()
        constants = set()
        visited = set()

        depth_first_search(target_vertex)

        # Remove duplicates while preserving order
        functions_order = list(dict.fromkeys(functions_order))

        # Sort inputs and constants
        sorted_user_inputs = {}
        sorted_constants = {}

        # Filter and include only the necessary user inputs
        for input_var in user_inputs:
            for key, schema in self.possible_user_inputs.items():
                if input_var in schema:
                    if key not in sorted_user_inputs:
                        sorted_user_inputs[key] = {}
                    sorted_user_inputs[key][input_var] = None

        # Filter and include only the necessary constants
        for const_var in constants:
            for key, schema in self.possible_constants.items():
                if const_var in schema:
                    if key not in sorted_constants:
                        sorted_constants[key] = {}
                    sorted_constants[key][const_var] = None
                elif key == "f_imb" and const_var in schema.index:
                    if key not in sorted_constants:
                        sorted_constants[key] = schema 

        if report:
            print_report(
                target_variable, functions_order, sorted_user_inputs, 
                sorted_constants
                )

        return {
            "functions_order": functions_order,
            "user_inputs": sorted_user_inputs,
            "constants": sorted_constants
        }

    def create_function(self, target_variable: str) -> Callable[..., Any]:
        """
        Create a dynamically generated function to calculate the target variable.

        This method generates and returns a function that can be used to calculate 
        the specified target variable. The returned function will have all required 
        arguments and the necessary logic to calculate the target variable based on 
        the DAG structure.

        Args:
            target_variable: The name of the target variable being calculated.

        Returns:
            Callable: A function that calculates the target variable.
        """
        def create_docstring(
            target_variable: str, 
            arg_names: List[str], 
            user_inputs: Dict[str, Dict[str, Optional[str]]], 
            constants: Dict[str, Dict[str, Optional[str]]], 
            functions_order: List[str], 
            generated_func_return: str
        ) -> str:
            """
            Generate a docstring for a dynamically generated function.

            This function creates a docstring that describes the dynamically 
            generated function used to calculate the specified target variable. 
            The docstring includes information about the function's arguments, 
            the order of function calls, and the return value.

            Args:
                target_variable: The name of the target variable being calculated.
                arg_names: A list of argument names required by the function.
                user_inputs: A dictionary of required user input structures and 
                    their respective fields.
                constants: A dictionary of required constants and their respective 
                    fields, excluding special cases like "aa_list".
                functions_order: A list of functions in the order they should be called.
                generated_func_return: The name of the variable that the function 
                    returns.

            Returns:
                A string representing the docstring for the generated function.
            """
            docstring = (
                f'"""Dynamically generated function to calculate '
                f'{target_variable}.\n\n'
                )
            docstring += 'Arguments:\n'
            for arg in arg_names:
                docstring += (
                    f'    {arg} (dict): A dictionary containing the following '
                    'keys:\n'
                    )
                if arg in user_inputs:
                    for key in user_inputs[arg].keys():
                        docstring += f'            - {key}\n'
                elif arg in constants and arg != "aa_list":
                    for key in constants[arg].keys():
                        docstring += f'            - {key}\n'
            
            docstring += '\nOrder of function calls:\n'
            for i, func in enumerate(functions_order, 1):
                docstring += f'    {i}. {func}\n'

            docstring += '\nReturns:\n'
            docstring += (
                f'    {generated_func_return} (float): The calculated value for '
                f'{target_variable}.\n')
            docstring += '"""'
            return docstring


        def check_for_constants(
            functions_order: List[str], 
            constants: Dict[str, Dict[str, str]]
        ) -> bool:
            """
            Check for 'coeff_dict' and 'aa_list' in function arguments.

            This function checks if 'coeff_dict' and 'aa_list' are required by any 
            function in the given order. For some wrapper functions, such as 
            calculate_Dt_DMIn, 'coeff_dict' might be passed to other functions 
            without accessing any keys, so it won't appear in the Constants list 
            in dag_data. The function returns a boolean indicating whether 
            'aa_list' is required.

            Args:
                functions_order: A list of function names to check.
                constants: A dictionary of constants already identified.

            Returns:
                bool: True if 'aa_list' is required by any function, False otherwise.
            """
            aa_list_required = False
            for function_name in functions_order:
                func = getattr(nd, function_name)
                func_args = inspect.signature(func).parameters.keys()
                if "coeff_dict" in func_args and "coeff_dict" not in constants:
                    constants["coeff_dict"] = {} 
                if "aa_list" in func_args and "aa_list" not in constants:
                    aa_list_required = True
            return aa_list_required


        def _identify_dict_args(
            dag_data: pd.DataFrame, 
            functions_order: List[str]
        ) -> Dict[str, Dict[str, List[Any]]]:
            """
            Identify functions that require dictionaries as arguments.

            This function examines the functions in the given order to identify 
            which ones require dictionaries as arguments, excluding 'coeff_dict'. 
            It uses type annotations to determine if an argument is a dictionary 
            and retrieves the associated arguments from the DAG data.

            Args:
                dag_data: A DataFrame containing information about the DAG, 
                    including function names and their arguments.
                functions_order: A list of function names to check.

            Returns:
                A dictionary where each key is a function name, and each value 
                is another dictionary that maps the dictionary argument name 
                to a list of required arguments.
            """
            dict_inputs = {}

            for function_name in functions_order:
                func = getattr(nd, function_name)
                func_signature = inspect.signature(func)
                dict_inputs[function_name] = {}
               
                # Check each parameter's type annotation to see if it is a dict (excluding 'coeff_dict')
                for param_name, param in func_signature.parameters.items():
                    annotation = param.annotation
                    if (param_name != "coeff_dict" and 
                       (annotation is dict or 
                        annotation is Dict)
                       ):
                        dag_row = dag_data[dag_data['Function'] == function_name]
                        
                        if not dag_row.empty:
                            arguments_list = dag_row['Arguments'].values[0]
                            dict_inputs[function_name][param_name] = arguments_list
                        else:
                            print(
                                f"Warning: No entry found in DAG data for "
                                f"function '{function_name}'"
                                )
            
                if not dict_inputs[function_name]:
                    del dict_inputs[function_name]

            return dict_inputs


        # Call get_calculation_order to get requirements
        requirements = self.get_calculation_order(target_variable, report=False)
        functions_order = requirements["functions_order"]
        user_inputs = requirements["user_inputs"]
        constants = requirements["constants"]

        dict_inputs = _identify_dict_args(self.dag_data, functions_order)
        aa_list_required = check_for_constants(functions_order, constants)

        arg_names = sorted(list(user_inputs.keys()) + list(constants.keys()))

        func_name_to_result_name = {}
        for func in functions_order:
            result_name = (
                self.dag_data[self.dag_data['Function'] == func]['Name'].values[0]
                )
            func_name_to_result_name[func] = result_name
        
        # Define the function dynamically using exec
        func_args = ", ".join(arg_names)
        generated_func_name = f"wrapper_{target_variable}"
        #NOTE Update return
        generated_func_return = list(func_name_to_result_name.values())[-1]
        docstring = create_docstring(
            target_variable, arg_names, user_inputs, constants, functions_order,
            generated_func_return
            )
        func_body = f"""
def {generated_func_name}({func_args}):
    {docstring}
    dict_inputs = {dict_inputs}    
    if {aa_list_required}:
        aa_list = {self.aa_list.copy()}

    # Step 1: Unpack nested values from user inputs and constants
    for input_dict_name, input_dict in {user_inputs}.items():
        for key in input_dict:
            locals()[key] = locals()[input_dict_name].get(key, None)
    
    for const_dict_name in {constants}.keys():
        locals()[const_dict_name] = const_dict_name

    # Step 2: Call each function in order and store the results
    for function_name in {functions_order}:
        func = getattr(nd, function_name)
        func_args = inspect.signature(func).parameters.keys()

        # Create dictionary if requried
        if function_name in dict_inputs.keys():
            for dict_name, required_keys in dict_inputs[function_name].items():
                locals()[dict_name] = {{
                    key: locals()[key] for key in required_keys
                    }}
                
        # Resolve arguments for the function call
        func_call_args = {{
            arg: locals()[arg] for arg in func_args if arg in locals()
            }}

        # Call the function and store the result with the correct name
        result = func(**func_call_args)
        result_name = {func_name_to_result_name}[function_name]
        locals()[result_name] = result
   
    exclude_keys = {
        'dict_inputs', 'input_dict_name', 'input_dict', 'function_name', 
        'func', 'func_args', 'func_call_args', 'result', 'result_name', 
        'exclude_keys'
        }
    locals_dict = {{
        key: value for key, value in locals().items() if key not in exclude_keys
    }}
    model_output = output.ModelOutput(locals_input=locals_dict)
    return model_output
"""
        # Execute the dynamic function definition
        exec_namespace = {}
        exec(func_body, globals(), exec_namespace)
        generated_function = exec_namespace[generated_func_name]

        return generated_function
    


"""Protein metabolism and utilization calculations.

This module includes functions to estimate the utilization, synthesis, and 
degradation of protein.
"""

import numpy as np
import pandas as pd


def calculate_f_mPrt_max(An_305RHA_MlkTP: float, coeff_dict: dict) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {'K_305RHA_MlkTP': 1.0}

    calculate_f_mPrt_max(
        An_305RHA_MlkTP = 400.0, coeff_dict = coeff_dict
    )
    ```
    """
    # Line 2116, 280kg RHA ~ 930 g mlk NP/d herd average
    f_mPrt_max = 1 + coeff_dict['K_305RHA_MlkTP'] * (An_305RHA_MlkTP / 280 - 1)
    return f_mPrt_max


def calculate_Du_MiCP_g(Du_MiN_g: float) -> float:
    Du_MiCP_g = Du_MiN_g * 6.25  # Line 1163
    return Du_MiCP_g


def calculate_Du_MiTP_g(Du_MiCP_g: float, coeff_dict: dict) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {'fMiTP_MiCP': 0.824}

    calculate_Du_MiTP_g(
        Du_MiCP_g = 100.0, coeff_dict = coeff_dict
    )
    ```
    """
    Du_MiTP_g = coeff_dict['fMiTP_MiCP'] * Du_MiCP_g  # Line 1166
    return Du_MiTP_g


def calculate_Scrf_CP_g(An_StatePhys: str, An_BW: float) -> float:
    """
    Scrf_CP_g: Scurf CP, g
    """
    if An_StatePhys == "Calf":
        Scrf_CP_g = 0.219 * An_BW**0.60  # Line 1965
    else:
        Scrf_CP_g = 0.20 * An_BW**0.60  # Line 1964
    return Scrf_CP_g


def calculate_Scrf_NP_g(Scrf_CP_g: float, coeff_dict: dict) -> float:
    """
    Scrf_NP_g: Scurf Net Protein, g
    
    Examples
    --------
    ```
    coeff_dict = {'Body_NP_CP': 0.86}

    calculate_Scrf_NP_g(
        Scrf_CP_g = 150.0, coeff_dict = coeff_dict
    )
    ```
    """
    Scrf_NP_g = Scrf_CP_g * coeff_dict['Body_NP_CP']  # Line 1966
    return Scrf_NP_g


def calculate_Scrf_MPUse_g_Trg(
    An_StatePhys: str, 
    Scrf_CP_g: float,
    Scrf_NP_g: float, 
    Km_MP_NP_Trg: float
) -> float:
    """
    Scrf_MPUse_g_Trg: Scurf Metabolizable protein, g
    """
    if An_StatePhys in ["Calf", "Heifer"]:
        Scrf_MPUse_g_Trg = Scrf_CP_g / Km_MP_NP_Trg  
        # calves and heifers are CP based., Line 2671
    else:
        Scrf_MPUse_g_Trg = Scrf_NP_g / Km_MP_NP_Trg  # Line 2670
    return Scrf_MPUse_g_Trg


def calculate_Scrf_NP(Scrf_NP_g: float) -> float:
    """
    Scrf_NP: Scurf net protein (kg/d)
    """
    Scrf_NP = Scrf_NP_g * 0.001  # Line 1967
    return Scrf_NP


def calculate_Scrf_N_g(Scrf_CP_g: float) -> float:
    """
    Scrf_N_g: Scurf N (g/d)
    """
    Scrf_N_g = Scrf_CP_g * 0.16  # Line 1968
    return Scrf_N_g


def calculate_Scrf_AA_TP(aa_list: list, coeff_dict: dict) -> np.ndarray:
    Scrf_AA_TP = np.array([coeff_dict[f"Scrf_{aa}_TP"] for aa in aa_list])
    return Scrf_AA_TP


def calculate_Scrf_AA_g(Scrf_NP_g: float, Scrf_AA_TP: np.ndarray) -> np.ndarray:
    """
    Scrf_AA_g: AA in scurf (g/d)
    """
    Scrf_AA_g = Scrf_NP_g * Scrf_AA_TP / 100  # Lines 1969-1978
    return Scrf_AA_g


def calculate_ScrfAA_AbsAA(
    Scrf_AA_g: pd.Series,
    Abs_AA_g: pd.Series
) -> np.array:
    """
    ScrfAA_AbsAA: Scurf aa as a fraction of absorbed aa
    """
    ScrfAA_AbsAA = Scrf_AA_g / Abs_AA_g  # Line 1981-1990
    return ScrfAA_AbsAA


def calculate_An_CPxprt_g(
    Scrf_CP_g: float, 
    Fe_CPend_g: float, 
    Mlk_CP_g: float,
    Body_CPgain_g: float
) -> float:
    """
    An_CPxprt_g: CP used for export protein (g/d)
    """
    An_CPxprt_g = Scrf_CP_g + Fe_CPend_g + Mlk_CP_g + Body_CPgain_g  
    # Initially defined only as true export protein, but it has migrated to 
    # include other prod proteins, Line 2525
    return An_CPxprt_g


def calculate_An_NPxprt_g(
    Scrf_NP_g: float, 
    Fe_NPend_g: float, 
    Mlk_NP_g: float,
    Body_NPgain_g: float
) -> float:
    """
    An_NPxprt_g: NP used for export protein (g/d)
    """
    An_NPxprt_g = Scrf_NP_g + Fe_NPend_g + Mlk_NP_g + Body_NPgain_g  
    # Should have changed the name, Line 2526
    return An_NPxprt_g


def calculate_Trg_NPxprt_g(
    Scrf_NP_g: float, 
    Fe_NPend_g: float,
    Trg_Mlk_NP_g: float, 
    Body_NPgain_g: float
) -> float:
    """
    Trg_NPxprt_g: NP used for export protein (g/d)
    """
    Trg_NPxprt_g = Scrf_NP_g + Fe_NPend_g + Trg_Mlk_NP_g + Body_NPgain_g  
    # Shouldn't these also include Gest??, Line 2527
    return Trg_NPxprt_g


def calculate_An_CPprod_g(
    Mlk_CP_g: float, 
    Gest_NCPgain_g: float,
    Body_CPgain_g: float
) -> float:
    """
    An_CPprod_g: CP use for production (g/d)
    """
    An_CPprod_g = Mlk_CP_g + Gest_NCPgain_g + Body_CPgain_g  
    # CP use for production. Be careful not to double count Gain. Line 2529
    return An_CPprod_g


def calculate_An_NPprod_g(
    Mlk_NP_g: float, 
    Gest_NPgain_g: float,
    Body_NPgain_g: float
) -> float:
    """
    An_NPprod_g: NP use for production (g/d)
    """
    An_NPprod_g = Mlk_NP_g + Gest_NPgain_g + Body_NPgain_g  
    # NP use for production, Line 2530
    return An_NPprod_g


def calculate_Trg_NPprod_g(
    Trg_Mlk_NP_g: float, 
    Gest_NPgain_g: float,
    Body_NPgain_g: float
) -> float:
    """
    Trg_NPprod_g: NP used fpr production (g/d)
    """
    Trg_NPprod_g = Trg_Mlk_NP_g + Gest_NPgain_g + Body_NPgain_g  
    # NP needed for production target, Line 2531
    return Trg_NPprod_g


def calculate_An_NPprod_MPIn(An_NPprod_g: float, An_MPIn_g: float) -> float:
    """
    An_NPprod_MPIn: NP used for produciton as fraction of metabolizable protein
    """
    An_NPprod_MPIn = An_NPprod_g / An_MPIn_g  # Line 2532
    return An_NPprod_MPIn


def calculate_Trg_NPuse_g(
    Scrf_NP_g: float, 
    Fe_NPend_g: float,
    Ur_NPend_g: float, 
    Trg_Mlk_NP_g: float,
    Body_NPgain_g: float, 
    Gest_NPgain_g: float
) -> float:
    """
    Trg_NPuse_g: Target NP use (g/d)
    """
    Trg_NPuse_g = (Scrf_NP_g + Fe_NPend_g + Ur_NPend_g + 
                   Trg_Mlk_NP_g + Body_NPgain_g + Gest_NPgain_g) # Line 2535
    return Trg_NPuse_g


def calculate_An_NPuse_g(
    Scrf_NP_g: float, 
    Fe_NPend_g: float, 
    Ur_NPend_g: float,
    Mlk_NP_g: float, 
    Body_NPgain_g: float,
    Gest_NPgain_g: float
) -> float:
    """
    An_NPuse_g: NP use (g/d)
    """
    An_NPuse_g = (Scrf_NP_g + Fe_NPend_g + Ur_NPend_g + 
                  Mlk_NP_g + Body_NPgain_g + Gest_NPgain_g)  
    # includes only net use of true protein. Excludes non-protein maintenance
    # use, Line 2536
    return An_NPuse_g


def calculate_An_NCPuse_g(
    Scrf_CP_g: float, 
    Fe_CPend_g: float,
    Ur_NPend_g: float, 
    Mlk_CP_g: float,
    Body_CPgain_g: float, 
    Gest_NCPgain_g: float
) -> float:
    """
    An_NCPuse_g: Net CP use (g/d)
    """
    An_NCPuse_g = (Scrf_CP_g + Fe_CPend_g + Ur_NPend_g + 
                   Mlk_CP_g + Body_CPgain_g + Gest_NCPgain_g)  
    # Net CP use, Line 2537
    return An_NCPuse_g


def calculate_An_Nprod_g(
    Gest_NCPgain_g: float, 
    Body_CPgain_g: float,
    Mlk_CP_g: float
) -> float:
    """
    An_Nprod_g: N used for production (g/d)
    """
    An_Nprod_g = (Gest_NCPgain_g + Body_CPgain_g) / 6.25 + Mlk_CP_g / 6.34  
    # Line 2539
    return An_Nprod_g


def calculate_An_Nprod_NIn(An_Nprod_g: float, An_NIn_g: float) -> float:
    """
    An_Nprod_NIn: N used for production as fraction of N intake
    """
    An_Nprod_NIn = An_Nprod_g / An_NIn_g  # Line 2540
    return An_Nprod_NIn


def calculate_An_Nprod_DigNIn(An_Nprod_g: float, An_DigNtIn_g: float) -> float:
    """
    An_Nprod_DigNIn: N used for production as fraction of digestable N intake (g/d)
    """
    An_Nprod_DigNIn = An_Nprod_g / An_DigNtIn_g  # Line 2541
    return An_Nprod_DigNIn


def calculate_An_MPBal_g_Trg(An_MPIn_g: float, An_MPuse_g_Trg: float) -> float:
    """
    An_MPBal_g_Trg: Target MP balance (g/d)
    """
    An_MPBal_g_Trg = An_MPIn_g - An_MPuse_g_Trg  # Line 2700
    return An_MPBal_g_Trg


def calculate_Xprt_NP_MP_Trg(
    Scrf_NP_g: float, 
    Fe_NPend_g: float,
    Trg_Mlk_NP_g: float, 
    Body_NPgain_g: float,
    An_MPIn_g: float, 
    Ur_NPend_g: float,
    Gest_MPUse_g_Trg: float
) -> float:
    """
    Xprt_NP_MP_Trg: Export NP to MP efficiency
    """
    Xprt_NP_MP_Trg = ((Scrf_NP_g + Fe_NPend_g + Trg_Mlk_NP_g + Body_NPgain_g) / 
                      (An_MPIn_g - Ur_NPend_g - Gest_MPUse_g_Trg))
    # Predicted An_NP_MP using Target Milk NP, g/g, Line 2701
    # The above excludes Ur_NPend and Gest_NPgain from the denominator and the 
    # numerator as that is how Helene and Roger derived target efficiencies.
    # Seems incorrect unless one does not consider Urinary endogenous as NP 
    # which is inconsistent with the definition. Not a true, total MP efficiency
    return Xprt_NP_MP_Trg


def calculate_Xprt_NP_MP(
    Scrf_NP_g: float, 
    Fe_NPend_g: float, 
    Mlk_NP_g: float,
    Body_NPgain_g: float, 
    An_MPIn_g: float,
    Ur_NPend_g: float, 
    Gest_MPUse_g_Trg: float
) -> float:
    """
    Xprt_NP_MP: Export efficiency NP to MP
    """
    Xprt_NP_MP = ((Scrf_NP_g + Fe_NPend_g + Mlk_NP_g + Body_NPgain_g) / 
                  (An_MPIn_g - Ur_NPend_g - Gest_MPUse_g_Trg))  # Line 2720
    return Xprt_NP_MP


def calculate_Km_MP_NP(An_StatePhys: str, Xprt_NP_MP: float) -> float:
    """
    Km_MP_NP: MP to NP conversion efficiency
    """
    # MP to NP conversion efficiency assuming 100% efficiency of Ur_NPend.  
    # Should not be named as a K as it is not a constant. 
    # This is apparent efficiency. Assumed equal to efficiency for total
    # proteins except for heifers.
    if An_StatePhys == "Heifer":  # Line 2722
        Km_MP_NP = 0.69
    else:
        Km_MP_NP = Xprt_NP_MP
    return Km_MP_NP


def calculate_Kl_MP_NP(Xprt_NP_MP: float) -> float:
    """
    Kl_MP_NP: MP to NP conversion efficiency
    """
    Kl_MP_NP = Xprt_NP_MP  # Line 2723
    return Kl_MP_NP


def calculate_Scrf_MPUse_g(Scrf_NP_g: float, Km_MP_NP: float) -> float:
    """
    Scrf_MPUse_g: Endogenous MP in scurf (g/d)
    """
    Scrf_MPUse_g = Scrf_NP_g / Km_MP_NP  # Line 2727
    return Scrf_MPUse_g


def calculate_An_MPuse_g(
    Fe_MPendUse_g: float, 
    Scrf_MPUse_g: float, 
    Ur_MPendUse_g: float, 
    Body_MPUse_g_Trg: float, 
    Gest_MPUse_g_Trg: float, 
    Mlk_MPUse_g: float,
) -> float:
    """
    An_MPuse_g: Total MP use (g/d)
    """
    An_MPuse_g = (Fe_MPendUse_g + Scrf_MPUse_g + Ur_MPendUse_g + 
                  Body_MPUse_g_Trg + Gest_MPUse_g_Trg + Mlk_MPUse_g)
    return An_MPuse_g


def calculate_An_MPuse(An_MPuse_g: float):
    """
    An_MPuse: Total MP use (kg/d)
    """
    An_MPuse = An_MPuse_g / 1000  # Line 2730
    return An_MPuse


def calculate_An_MPBal_g(An_MPIn_g: float, An_MPuse_g: float) -> float:
    """
    An_MPBal_g: Metabolizable protein balance (g/d)
    """
    An_MPBal_g = An_MPIn_g - An_MPuse_g  
    # This will always be 0 given how the efficiencies are calculated. 
    # Not informative, Line 2731
    return An_MPBal_g


def calculate_An_MP_NP(An_NPuse_g: float, An_MPuse_g: float) -> float:
    """
    An_MP_NP: Conversion of MP to NP 
    """
    An_MP_NP = An_NPuse_g / An_MPuse_g  #Total MP Efficiency, g/g, Line 2734
    return An_MP_NP


def calculate_An_NPxprt_MP(
    An_NPuse_g: float, 
    Ur_NPend_g: float,
    Gest_NPuse_g: float, 
    An_MPIn_g: float,
    Gest_MPUse_g_Trg: float
) -> float:
    """
    An_NPxprt_MP: NP to MP conversion efficiency
    """
    An_NPxprt_MP = ((An_NPuse_g - Ur_NPend_g - Gest_NPuse_g) / 
                    (An_MPIn_g - Ur_NPend_g - Gest_MPUse_g_Trg))
    # g/g. For comparison to target efficiency, but a duplicate of Xprt_NP_MP
    # Line 2735
    return An_NPxprt_MP


def calculate_An_CP_NP(An_NPuse_g: float, An_CPIn: float) -> float:
    """
    An_CP_NP: CP to NP conversion efficiency 
    """
    An_CP_NP = An_NPuse_g / (An_CPIn * 1000)
    # Total CP efficiency, g/g, Line 2736
    return An_CP_NP


def calculate_An_NPBal_g(
    An_MPIn_g: float, 
    An_MP_NP: float,
    An_NPuse_g: float
) -> float:
    """
    An_NPBal_g: Net protein balance (g/d)
    """
    An_NPBal_g = An_MPIn_g * An_MP_NP - An_NPuse_g  
    # Also always 0, thus non-informative, Line 2738
    return An_NPBal_g


def calculate_An_NPBal(An_NPBal_g: float) -> float:
    """
    An_NPBal: Net protein balance (kg/d)
    """
    An_NPBal = An_NPBal_g / 1000  # Line 2739
    return An_NPBal
